from typing import TypeVar, Mapping, List, TypedDict, Optional, Sequence
from pathlib import Path

from omnibenchmark.utils.user_input_checks import check_name_matching, flatten
from omnibenchmark.utils.auto_input import (
    get_input_files_from_prefix,
    get_parameter_from_dataset,
)
from omnibenchmark.utils.exceptions import InputError
from omnibenchmark.management.data_commands import update_datasets_by_keyword
from omnibenchmark.management.parameter_checks import (
    filter_parameter,
    get_all_parameter_combinations,
    filter_parameter_combinations,
)

PathLike = TypeVar("PathLike", str, Path, None)


class OmniInput:
    def __init__(
        self,
        names: List[str],
        prefix: Optional[Mapping[str, List[str]]] = None,
        input_files: Optional[Mapping[str, Mapping[str, str]]] = None,
        keyword: Optional[List[str]] = None,
        default: Optional[str] = None,
    ):
        self.names = names
        self.prefix = prefix
        self.input_files = input_files
        self.keyword = keyword
        self.default = default

        if self.input_files is None:

            if self.prefix is None or self.keyword is None:
                raise InputError(
                    f"Missing input file definition. \n"
                    f"Please specify explicit input files or define prefixes AND dataset keyword(s) to automatically identify them."
                )

            check_name_matching(self.names, self.prefix.keys())
            self.input_files = get_input_files_from_prefix(self.prefix, self.keyword)

        if len(self.input_files) < 1:
            self.input_files = None
            print(
                f"WARNING: No input files in the current project detected.\n"
                f"Run OmniObject.update_object() to update and import input datasets.\n"
                f"Otherwise check the specified prefixes: {self.prefix}"
            )

        if self.input_files is not None:
            check_name_matching(
                self.names, flatten([pre.keys() for pre in self.input_files.values()])
            )

            if self.default is None:
                self.default = next(iter(self.input_files.items()))[0]

            if self.default not in self.input_files.keys():
                raise InputError(
                    f"Input default {self.default} not found. Must have corresponding input files."
                )

            self.default_files = self.input_files[self.default]

    def update_inputs(
        self, orchestrator: str, query_url: str, data_url: str, gitlab_url: str
    ):
        if self.keyword is not None:
            for key in self.keyword:
                update_datasets_by_keyword(
                    keyword=key,
                    o_url=orchestrator,
                    query_url=query_url,
                    data_url=data_url,
                    gitlab_url=gitlab_url,
                )
            if self.prefix is not None:
                check_name_matching(self.names, self.prefix.keys())
                self.input_files = get_input_files_from_prefix(
                    self.prefix, self.keyword
                )
                check_name_matching(
                    self.names,
                    flatten([pre.keys() for pre in self.input_files.values()]),
                )
            if self.input_files is not None:
                if self.default is None:
                    self.default = next(iter(self.input_files.items()))[0]
        #return self


class OmniParameter:
    def __init__(
        self,
        names: List[str],
        values: Optional[Mapping[str, List]] = None,
        default: Optional[Mapping[str, str]] = None,
        keyword: Optional[List[str]] = None,
        filter: Optional[Mapping[str, str]] = None,
        combinations: Optional[List[Mapping[str, str]]] = None,
    ):
        self.names = names
        self.values = values
        self.default = default
        self.keyword = keyword
        self.filter = filter
        self.combinations = combinations

        if self.values is None and self.keyword is not None:
            self.values = get_parameter_from_dataset(self.names, self.keyword)

        if self.values is not None:
            check_name_matching(self.names, self.values.keys())
            self.values = filter_parameter(self.values, self.filter)

            if self.combinations is None:
                self.combinations = get_all_parameter_combinations(self.values)

        if self.combinations is not None:
            check_name_matching(
                self.names, flatten([comb.keys() for comb in self.combinations])
            )
            self.combinations = filter_parameter_combinations(
                self.combinations, self.filter
            )
            if self.default is None:
                self.default = self.combinations[0]

    def update_parameter(
        self, orchestrator: str, query_url: str, data_url: str, gitlab_url: str
    ):
        if self.keyword is not None:
            for key in self.keyword:
                update_datasets_by_keyword(
                    keyword=key,
                    o_url=orchestrator,
                    query_url=query_url,
                    data_url=data_url,
                    gitlab_url=gitlab_url,
                )

            self.values = get_parameter_from_dataset(self.names, self.keyword)
            check_name_matching(self.names, self.values.keys())
            self.values = filter_parameter(self.values, self.filter)
            self.combinations = get_all_parameter_combinations(self.values)

            if self.combinations is not None:
                check_name_matching(
                    self.names, flatten([comb.keys() for comb in self.combinations])
                )
                self.combinations = filter_parameter_combinations(
                    self.combinations, self.filter
                )
                if self.default is None:
                    self.default = self.combinations[0]
        #return self


class OutMapping(TypedDict):
    output_files: Optional[Mapping]
    input_files: Optional[Mapping]
    parameter: Optional[Mapping]



from omnibenchmark.utils import auto_run as omni
from renku.domain_model.workflow.parameter import CommandParameter, CommandInput


# Test map_plan_names
def test_map_plan_names_empty_plan(mock_plan):
    assert omni.map_plan_names(mock_plan) == {}


def test_map_plan_names_with_all(mock_plan, mock_input, mock_output, mock_parameter):
    mock_plan.inputs = [mock_input]
    mock_plan.outputs = [mock_output]
    mock_plan.parameters = [mock_parameter]
    assert omni.map_plan_names(mock_plan) == {
        "test_input-1": "test/input",
        "test_output-1": "test/output",
        "test_param-1": "test/param",
    }


# Test get_file_type_dict
def test_get_file_type_dict_works(mock_out_mapping):
    assert omni.get_file_type_dict(mock_out_mapping) == {
        "out_file1": "path/to/out1",
        "out_file2": "path/to/out2",
        "in_file1": "path/to/in1",
        "param1": "str_value",
        "param2": 100,
    }


# Test get_file_name_dict
def test_get_file_name_dict_multiple_mapping(mock_file_list):
    assert omni.get_file_name_dict(mock_file_list) == {
        "path/to/dim/red": "dim_red_file",
        "path/to/count": "count_file",
        "0": "param3",
        "test": "param2",
    }


# Test map_plan_names_by_prefix
def test_map_plan_names_by_prefix_with_double(mock_plan):
    mock_param1 = CommandParameter(
        id="AAA", default_value="0", name="param3-3", prefix="param3"
    )
    mock_param2 = CommandParameter(
        id="AAA", default_value="0", name="param1-1", prefix="param1"
    )
    mock_plan.parameters = [mock_param1, mock_param2]
    plan_dict = omni.map_plan_names(mock_plan)
    assert omni.map_plan_names_by_prefix(plan=mock_plan, plan_dict=plan_dict) == {
        "param1": "param1-1",
        "param3": "param3-3",
    }


def test_map_plan_names_by_prefix_without_double(mock_plan, mock_input):
    mock_plan.inputs = [mock_input]
    plan_dict = omni.map_plan_names(mock_plan)
    assert omni.map_plan_names_by_prefix(plan=mock_plan, plan_dict=plan_dict) == {}


def test_map_plan_names_by_prefix_with_double_none(mock_plan):
    mock_param1 = CommandParameter(
        id="AAA", default_value="0", name="param3-3", prefix="param3"
    )
    mock_param2 = CommandParameter(id="AAA", default_value="0", name="param1-1")
    mock_plan.parameters = [mock_param1, mock_param2]
    plan_dict = omni.map_plan_names(mock_plan)
    assert omni.map_plan_names_by_prefix(plan=mock_plan, plan_dict=plan_dict) == {
        "param3": "param3-3"
    }


# Test map_plan_names_file_types
def test_map_plan_names_file_types_duplicates(mock_plan, mock_file_list, mock_input):
    mock_input.default_value = "path/to/dim/red"
    mock_input.name = "dim_red_file-1"
    mock_in2 = CommandInput(
        id="AAA", default_value="path/to/count", name="count_file-1"
    )
    mock_param1 = CommandParameter(
        id="AAA", default_value="0", name="param3-3", prefix="param3"
    )
    mock_param2 = CommandParameter(
        id="AAA", default_value="0", name="param1-1", prefix="param1"
    )
    mock_plan.parameters = [mock_param1, mock_param2]
    mock_plan.inputs = [mock_input, mock_in2]
    mock_file_dict = omni.get_file_name_dict(file_mapping=mock_file_list)
    assert omni.map_plan_names_file_types(plan=mock_plan, file_dict=mock_file_dict) == {
        "count_file": "count_file-1",
        "dim_red_file": "dim_red_file-1",
        "param1": "param1-1",
        "param3": "param3-3",
    }


def test_map_plan_names_file_types_no_match(mock_plan, mock_file_list):
    mock_file_dict = omni.get_file_name_dict(file_mapping=mock_file_list)
    assert (
        omni.map_plan_names_file_types(plan=mock_plan, file_dict=mock_file_dict) == {}
    )


def test_map_plan_names_file_types_not_in_list(mock_plan, mock_file_list):
    mock_in = CommandInput(
        id="AAA", default_value="new/path/to/count", name="new_count_file-1"
    )
    mock_plan.inputs = [mock_in]
    mock_file_dict = omni.get_file_name_dict(file_mapping=mock_file_list)
    assert (
        omni.map_plan_names_file_types(plan=mock_plan, file_dict=mock_file_dict) == {}
    )


# Test get_file_mapping_from_out_files
def test_get_file_mapping_from_out_files_match(mock_out_mapping):
    assert omni.get_file_mapping_from_out_files(
        out_files=["path/to/out1"], file_mapping=[mock_out_mapping]
    ) == [mock_out_mapping]


def test_get_file_mapping_from_out_files_no_match(mock_out_mapping):
    assert (
        omni.get_file_mapping_from_out_files(
            out_files=["new/path/to/out1"], file_mapping=[mock_out_mapping]
        )
        == []
    )


def test_get_file_mapping_from_out_files_no_mapping():
    assert omni.get_file_mapping_from_out_files(out_files=["new/path/to/out1"]) == []


# Test get_file_list_from_out_mapping
def test_get_file_list_from_out_mapping_one(mock_out_mapping):
    assert omni.get_file_list_from_out_mapping([mock_out_mapping]) == [
        "path/to/out1",
        "path/to/out2",
        "path/to/in1",
        "str_value",
        100,
    ]


def test_get_file_list_from_out_mapping_two(mock_out_mapping):
    assert omni.get_file_list_from_out_mapping(
        [mock_out_mapping, mock_out_mapping]
    ) == [
        "path/to/out1",
        "path/to/out2",
        "path/to/in1",
        "str_value",
        100,
        "path/to/out1",
        "path/to/out2",
        "path/to/in1",
        "str_value",
        100,
    ]


# Test get_file_list_from_out_mapping
def test_get_file_list_from_out_mapping_none():
    assert omni.get_file_list_from_out_mapping([]) == []


def test_get_file_list_from_out_mapping_half_empty(mock_out_mapping):
    mock_out_mapping["output_files"] = {}
    assert omni.get_file_list_from_out_mapping([mock_out_mapping]) == [
        "path/to/in1",
        "str_value",
        100,
    ]



from __future__ import annotations
import logging
from typing import Any, TypeVar, Optional, List, Union, Dict
from pathlib import Path
from renku.command.dataset import create_dataset_command
from renku.domain_model.dataset import Dataset as RenkuDataSet
import omnibenchmark.management.general_checks
from omnibenchmark.management.data_checks import dataset_name_exist, renku_dataset_exist
from renku.command.dataset import (
    import_dataset_command,
    update_datasets_command,
    add_to_dataset_command,
)
from renku.core import errors
from omnibenchmark.utils.exceptions import ProjectError

PathLike = TypeVar("PathLike", str, Path, None)
logger = logging.getLogger("omnibenchmark.renku_commands")


def renku_dataset_create(
    name: str,
    kg_url: str,
    title: Optional[str] = None,
    description: Optional[str] = None,
    creators: Optional[List[str]] = None,
    meta_data: Optional[Dict[str, Any]] = None,
    keyword: Optional[List[str]] = None,
) -> RenkuDataSet:
    """Generate an empty renku dataset in the current project. Works in a renku project only.

    Args:
        name ([str]): Dataset name
        title (str, optional): Dataset title. Defaults to None.
        description (str, optional): Description of the dataset. Defaults to None.
        creators (str, optional): Dataset creator. Defaults to None, which will be the github account activated for that renku project.
        metadata (PathLike, optional): Path to json or yaml file containing meta data information. Defaults to None.
        keyword (list[str], optional): List of keywords to be associated with the dataset. Defaults to None.

    Returns:
        RenkuDataSet: An object of class Dataset from renku.core.models.dataset
    """
    if not omnibenchmark.management.general_checks.is_renku_project():
        raise ProjectError(
            "Directory is not a renku project.\n"
            "Make sure you are in the correct context.\n"
            "No dataset was created."
        )

    if renku_dataset_exist(name):
        print(f"Dataset {name} already exists in this repository.")
        return

    if dataset_name_exist(name, kg_url):
        print(f"Dataset {name} already taken. Please use a different name.")
        return

    result = (
        create_dataset_command()
        .build()
        .execute(
            name=name,
            title=title,
            description=description,
            creators=creators,
            keywords=keyword,
            custom_metadata=meta_data,
        )
    )
    print(
        f"Renku dataset with name {name} and the following keywords {keyword} was generated."
    )
    return result.output


def renku_dataset_import(
    uri: str, name: str = None, extract: bool = False, yes: bool = True
):
    if not omnibenchmark.management.general_checks.is_renku_project():
        raise ProjectError(
            "Directory is not a renku project.\n"
            "Make sure you are in the correct context.\n"
            "No dataset was imported."
        )

    if name is not None:
        if renku_dataset_exist(name):
            logger.info(f"Dataset {name} already exists in this repository.")
            return

    result = (
        import_dataset_command()
        .build()
        .execute(uri=uri, name=name, extract=extract, yes=yes)
    )

    return result.output


def renku_dataset_update(
    names: List[str],
    creators: Optional[List[str]] = None,
    include: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
    ref: Optional[str] = None,
    delete: bool = True,
    no_external: bool = True,
    update_all: bool = False,
    dry_run: bool = False,
):
    if not omnibenchmark.management.general_checks.is_renku_project():
        raise ProjectError(
            "Directory is not a renku project.\n"
            "Make sure you are in the correct context.\n"
            "No dataset was updated."
        )

    if (
        not update_all
        and len(names) == 0
        and include is None
        and exclude is None
        and not dry_run
    ):
        raise errors.ParameterError(
            "Either names, update_all, dry_run, or include/exclude should be specified"
        )

    if len(names) >= 1 and update_all:
        raise errors.ParameterError("Cannot pass dataset names with update_all")
    elif (include or exclude is not None) and update_all:
        raise errors.ParameterError("Cannot pass include/exclude with update_all")

    result = (
        update_datasets_command()
        .build()
        .execute(
            names=names,
            creators=creators,
            include=include,
            exclude=exclude,
            ref=ref,
            delete=delete,
            no_external=no_external,
            update_all=update_all,
            dry_run=dry_run,
        )
    )

    return result.output


def renku_add_to_dataset(
    urls: List[str],
    dataset_name: str,
    external: bool = False,
    force: bool = False,
    overwrite: bool = True,
    create: bool = False,
    sources: Optional[List[Union[str, Path]]] = None,
    destination: Optional[List[str]] = None,
    ref: Optional[str] = None,
) -> Optional[RenkuDataSet]:

    if not omnibenchmark.management.general_checks.is_renku_project():
        raise ProjectError(
            "Directory is not a renku project.\n"
            "Make sure you are in the correct context.\n"
            "No files were added."
        )

    result = (
        add_to_dataset_command()
        .build()
        .execute(
            urls=urls,
            dataset_name=dataset_name,
            external=external,
            force=force,
            overwrite=overwrite,
            create=create,
            sources=sources,
            destination=destination,
            ref=ref,
        )
    )
    return result



"""Commands related to import and update relevant datasets"""
from typing import List, Mapping, Any, Optional, Tuple
from renku.ui.api.models.dataset import Dataset
from omnibenchmark.renku_commands.datasets import (
    renku_dataset_import,
    renku_dataset_update,
    renku_add_to_dataset,
)
import requests
import os.path
import gitlab
from iteration_utilities import unique_everseen  # type: ignore

# Find datasets by string
def query_datasets_by_string(
    string: str, url: str = "https://renkulab.io/knowledge-graph/datasets?query="
) -> List[Mapping[Any, Any]]:
    query_url = url + string
    response = requests.get(query_url)
    return response.json()


# Find dataset by match of property
def query_datasets_by_property(
    string: str,
    url: str = "https://renkulab.io/knowledge-graph/datasets?query=",
    match_string: Optional[str] = None,
    property_name: str = "keywords",
) -> List[Mapping[Any, Any]]:
    if match_string is None:
        match_string = string
    all_data_json = query_datasets_by_string(string=string, url=url)
    matched_dataset: List = []
    for dataset in all_data_json:
        data_prop = (
            [dataset[property_name]]
            if not isinstance(dataset[property_name], list)
            else dataset[property_name]
        )
        if any(prop_val.lower() == match_string.lower() for prop_val in data_prop):
            matched_dataset.append(dataset)
    return matched_dataset


# Filter exisiting datasets
def filter_existing(data_json: List) -> Tuple[List, List]:
    update_list: List = []
    import_list: List = []
    datasets = Dataset.list()
    name_list = [dataset.name for dataset in datasets]
    for data in data_json:
        if data["name"] in name_list:
            update_list.append(data)
        else:
            import_list.append(data)
    return import_list, update_list


# Get dataset field by matching property
def get_field_by_dataset_property(
    string: str,
    url: str = "https://renkulab.io/knowledge-graph/datasets?query=",
    property_name: str = "keywords",
    field: str = "identifier",
    filter_ex: bool = False,
) -> Tuple[List[str], List[str]]:
    data_json = query_datasets_by_property(
        string=string, url=url, property_name=property_name
    )
    up_json: List = []
    if filter_ex:
        data_json, up_json = filter_existing(data_json=data_json)

    return (
        [dataset[field] for dataset in data_json],
        [dataset["name"] for dataset in up_json],
    )


# Get data info by dataset id
def get_data_info_by_id(
    id: str, url: str = "https://renkulab.io/knowledge-graph/datasets/"
) -> Mapping[Any, Any]:
    query_url = url + id
    response = requests.get(query_url)
    return response.json()


def find_dataset_linked_to_wflow(
    info_list: List[Mapping], top: int = 2, filter_pat: str = "meta"
) -> List[Mapping]:
    origin_info = []
    for info in info_list:
        data_files = info["hasPart"]
        data_file_list = [
            data_fi["atLocation"]
            for data_fi in data_files
            if filter_pat not in data_fi["atLocation"]
        ]
        for data_file in data_file_list[:top]:
            file_lineage = requests.get(
                info["project"]["_links"][0]["href"]
                + "/files/"
                + data_file
                + "/lineage"
            )
            if "message" not in file_lineage.json().keys():
                origin_info.append(info)
                return origin_info
    return origin_info


def filter_duplicated_names(info_list: List[Mapping]) -> List[Mapping]:
    info_list = list(unique_everseen(info_list))
    name_list = [info["name"] for info in info_list]
    dup_names = set(
        [info_name for info_name in name_list if name_list.count(info_name) > 1]
    )
    uni_list = [info for info in info_list if info["name"] not in dup_names]
    for dup in dup_names:
        dup_info = [info for info in info_list if info["name"] == dup]
        dup_project = set([info["project"]["_links"][0]["href"] for info in dup_info])
        if len(dup_project) == 1:
            all_datasets = requests.get(
                list(dup_project)[0] + "/datasets"
            )  # type:ignore
            all_ids = [dat["identifier"] for dat in all_datasets.json()]
            origin_info = [info for info in dup_info if info["identifier"] in all_ids]
        else:
            origin_info = find_dataset_linked_to_wflow(dup_info)
        if len(origin_info) == 0:
            print(
                f"WARNING: Could not identify origin of {dup}.\n"
                f"If you anyways want to use and import it, consider manual import:\n"
                f"renku dataset import DATA_URL"
            )
            continue
        uni_list.append(*origin_info)
    return uni_list


# Get origin dataset ids
def get_origin_dataset_ids(
    ids: List[str], url: str = "https://renkulab.io/knowledge-graph/datasets/"
) -> List[Mapping]:
    all_infos: List = []
    for data_id in ids:
        data_info = get_data_info_by_id(id=data_id, url=url)
        if "sameAs" in data_info.keys():
            if not data_info["sameAs"] == data_info["url"]:
                same_id = data_info["sameAs"].split("/")[
                    -1
                ]  ## Unstable string manipulation??
                data_info = get_data_info_by_id(id=same_id, url=url)
        if data_info not in all_infos and "name" in data_info.keys():
            all_infos.append(data_info)
    origin_infos = filter_duplicated_names(all_infos)
    return origin_infos


def get_project_info_from_url(project_url=str) -> Mapping[Any, Any]:
    response = requests.get(project_url)
    return response.json()


def check_orchestrator(
    data_info: Mapping, o_url: str, gitlab_url: str = "https://renkulab.io/gitlab"
) -> Optional[str]:
    project_url = data_info["project"]["_links"][0]["href"]
    project_info = get_project_info_from_url(project_url)
    o_info = get_project_info_from_url(o_url)
    renku_git = gitlab.Gitlab(gitlab_url)
    o_git = renku_git.projects.get(o_info["identifier"])
    success = o_git.pipelines.list(status="success")
    bridge_projects: List = []
    for pipe in success:
        bridge_projects.extend(
            [
                brid.downstream_pipeline["project_id"]
                for brid in pipe.bridges.list()
                if brid.status == "success"
            ]
        )
        if project_info["identifier"] in bridge_projects:
            return data_info["url"]
    return None


def get_data_url_by_keyword(
    keyword: str,
    o_url: str,
    filter_ex: bool = False,
    query_url: str = "https://renkulab.io/knowledge-graph/datasets?query=",
    data_url: str = "https://renkulab.io/knowledge-graph/datasets/",
    gitlab_url: str = "https://renkulab.io/gitlab",
) -> Tuple[List[str], List[str]]:
    all_ids, up_exist = get_field_by_dataset_property(
        string=keyword, url=query_url, filter_ex=filter_ex
    )
    if len(all_ids) + len(up_exist) < 1:
        print(f"WARNING:No datasets found with keyword {keyword}")
        return [], []
    origin_infos = get_origin_dataset_ids(ids=all_ids, url=data_url)
    if len(origin_infos) + len(up_exist) < 1:
        print(
            f"WARNING:Could not identify dataset sources. Please check each of {data_url}{all_ids} to make sure they are the intended source"
        )
        return [], []
    omni_ids = [
        check_orchestrator(data_info=info, o_url=o_url, gitlab_url=gitlab_url)
        for info in origin_infos
    ]
    omni_ids = list(filter(None, omni_ids))
    if len(omni_ids) < 1:
        if len(up_exist) < 1:
            print(
                f"WARNING:No dataset with keyword {keyword} was associated to an omnibenchmark at ${o_url}\n"
                f"Please check if you specified the correct benchmark/url \n"
            )
        else:
            print(f"No new dataset with keyword {keyword} was found.")
    return omni_ids, up_exist  # type:ignore


def update_datasets_by_keyword(
    keyword: str,
    o_url: str,
    filter_ex: bool = True,
    query_url: str = "https://renkulab.io/knowledge-graph/datasets?query=",
    data_url: str = "https://renkulab.io/knowledge-graph/datasets/",
    gitlab_url: str = "https://renkulab.io/gitlab",
):
    imp_ids, up_names = get_data_url_by_keyword(
        keyword=keyword,
        o_url=o_url,
        filter_ex=filter_ex,
        query_url=query_url,
        data_url=data_url,
        gitlab_url=gitlab_url,
    )
    for id in imp_ids:
        renku_dataset_import(uri=id)
    for nam in up_names:
        print(f"Updated dataset {nam}.")
        renku_dataset_update(names=[nam])


def update_dataset_files(urls: List[str], dataset_name: str):
    datasets = Dataset.list()
    name_list = [dataset.name for dataset in datasets]
    if dataset_name not in name_list:
        print(
            f"WARNING: Dataset {dataset_name} does not exist.\n"
            f"Files will not be added to any dataset."
        )
        return
    dataset = [data for data in datasets if data.name == dataset_name][0]
    valid_urls = [url for url in urls if os.path.isfile(url)]
    dataset_files = [fi.path for fi in dataset.files]
    add_urls = [url for url in valid_urls if url not in dataset_files]
    if len(add_urls) > 0:
        renku_add_to_dataset(urls=add_urls, dataset_name=dataset_name)
        print(f"Added the following files to {dataset_name}:\n {add_urls}")
    return



from typing import Mapping, Optional, List, Union, Callable
from string import Template
import itertools

from omnibenchmark.core.input_classes import OmniInput, OmniParameter, OutMapping
from omnibenchmark.utils.decorators import option_list, option_str


@option_str
def join_parameter(parameter: Optional[Mapping[str, str]] = None) -> str:

    all_params = "__".join(
        [
            "{}_{}".format(str(param_nam), param_val)
            for param_nam, param_val in zip(
                parameter.keys(), parameter.values()  # type: ignore
            )
        ]
    )
    return all_params


@option_str
def join_inputs_parameter(
    input: Optional[str] = None, parameter: Optional[Mapping[str, str]] = None
) -> str:
    param_str = join_parameter(parameter)
    try:
        if param_str == "":
            joined = "".join([input, param_str])  # type: ignore
        else:
            joined = "__".join([input, param_str])  # type: ignore
    except:
        joined = param_str
    finally:
        return joined


def get_out_names_from_input_params(
    output_end: Mapping[str, str],
    name: str,
    input: Optional[str] = None,
    parameter: Optional[Mapping[str, str]] = None,
    out_template: str = "data/${name}/${name}_${unique_values}_${out_name}.${out_end}",
    **kwargs,
) -> Mapping[str, str]:
    """Generates automatic output file names from inputs and parameters used, based on a template defined in out_template.

    Args:
        input (str): Specific input for this output. Should be potentially uniquely identifyable shortcuts, e.g, dataset names.
        output_end (Mapping[str, str]): Mapping of all output file types and their corresponding endings, e.g. {'norm_counts': 'mtx.gz'}.
        name (str): Name of the method, dataset, etc. the output belongs to.
        parameters (Optional[Mapping[str, str]], optional): Mapping of the parameter an their values used to generate these outputs. Defaults to None.
        out_template (str, optional): Template to generate names from. Defaults to 'data/${name}/${name}__${unique_values}${out_name}.${out_end}'.

    Returns:
        Mapping[str, str]: A mapping of output filetypes and generated names, e.g. '{norm_counts: "data/meth1/meth1__data1__norm_counts.mtx.gz"}'.
    """

    # check template
    valid_keys = ["unique_values", "name", "out_name", "out_end"]
    for arg in kwargs.keys():
        valid_keys.append(arg)
    temp_keys = [
        s[1] or s[2] for s in Template.pattern.findall(out_template) if s[1] or s[2]
    ]
    if any(temp not in valid_keys for temp in temp_keys):
        raise NameError(
            f"Invalid output filename template: {out_template}. Please use valid keys as in {valid_keys} or consider specifying output filenames explicitly."
        )
    unique_values = join_inputs_parameter(input, parameter)
    sub_dict = kwargs
    sub_dict["unique_values"] = unique_values
    sub_dict["name"] = name
    output_names = {}
    for out in output_end.keys():
        out_end = output_end[out]
        sub_dict["out_end"] = out_end
        sub_dict["out_name"] = out
        output_file = Template(out_template).substitute(**sub_dict)
        output_names[out] = output_file.replace("..", ".")

    return output_names


@option_list
def get_input_file_list(inputs: Optional[OmniInput]) -> List:
    return (
        list(inputs.input_files.keys()) if inputs.input_files is not None else []       # type: ignore
    )  


@option_list
def get_parameter_combinations(parameter: Optional[OmniParameter]) -> List:
    return (
        parameter.combinations if parameter.combinations is not None else []             # type: ignore
    )  


@option_list
def get_input_parameter_combinations(
    inputs: Optional[OmniInput] = None, parameter: Optional[OmniParameter] = None
) -> List:
    ins = get_input_file_list(inputs)
    para = get_parameter_combinations(parameter)
    out_list = []

    all_comb = list(itertools.product(ins, para))
    if len(all_comb) >= 1:
        for comb in all_comb:
            out_dict: OutMapping = {
                "output_files": None,
                "input_files": {comb[0]: inputs.input_files[comb[0]]},  # type: ignore
                "parameter": comb[1],
            }
            out_list.append(out_dict)
    else:
        for in_name in ins:
            out_dict_in: OutMapping = {
                "output_files": None,
                "input_files": {in_name: inputs.input_files[in_name]},  # type: ignore
                "parameter": None,
            }
            out_list.append(out_dict_in)
        for para_dict in para:
            out_dict_para: OutMapping = {
                "output_files": None,
                "input_files": None,
                "parameter": para_dict,
            }
            out_list.append(out_dict_para)

    return out_list


def get_template_vars_dependencies(
    comb: OutMapping, fun: Callable[..., Mapping], **kwargs
) -> Mapping:
    return fun(comb, **kwargs)


def get_all_output_combinations(
    name: str,
    output_end: Mapping[str, str],
    inputs: Optional[OmniInput] = None,
    parameter: Optional[OmniParameter] = None,
    out_template: str = "data/${name}/${name}_${unique_values}_${out_name}.${out_end}",
    template_fun: Optional[Callable[..., Mapping]] = None,
    **kwargs,
) -> List[OutMapping]:

    comb_list = get_input_parameter_combinations(inputs=inputs, parameter=parameter)
    out_list = []

    for comb in comb_list:
        if template_fun is not None:
            temp_vars = get_template_vars_dependencies(
                comb=comb, fun=template_fun, **kwargs
            )
        else:
            temp_vars = kwargs
        try:
            input_str = next(iter(comb["input_files"].keys()))
            comb["input_files"] = comb["input_files"][input_str]
        except:
            input_str = None
        finally:
            out_names = get_out_names_from_input_params(
                name=name,
                output_end=output_end,
                input=input_str,
                parameter=comb["parameter"],
                out_template=out_template,
                **temp_vars,
            )
        comb["output_files"] = out_names
        out_list.append(comb)

    if len(out_list) == 0:
        out_names = get_out_names_from_input_params(
            name=name, output_end=output_end, out_template=out_template
        )
        out_map: OutMapping = {
            "output_files": out_names,
            "input_files": None,
            "parameter": None,
        }
        out_list.append(out_map)
    return out_list

def convert_values_to_string(d: Mapping) -> Mapping:
    return {key: str(d[key]) for key in d.keys()}

@option_list
def get_default(omni_class: Optional[Union[OmniParameter, OmniInput]]):
    default_val = omni_class.default            # type: ignore
    if isinstance(default_val, dict):
        default_val = convert_values_to_string(default_val)
    return default_val  


@option_list
def get_default_input(omni_input: Optional[OmniInput]):
    def_input_nam = omni_input.default  # type: ignore
    if def_input_nam is not None:
        return convert_values_to_string(
            omni_input.input_files[def_input_nam]   # type: ignore
            ) 
    else:
        return []


def get_default_outputs(
    file_mapping: List[OutMapping],
    inputs: Optional[OmniInput] = None,
    parameter: Optional[OmniParameter] = None,
) -> Optional[Mapping]:
    def_input = get_default_input(inputs)
    def_param = get_default(parameter)
    def_out = None
    for out_dict in file_mapping:
        if (
            out_dict["input_files"] is not None
            and out_dict["input_files"] == def_input
            and out_dict["parameter"] == def_param
        ):
            def_out = out_dict["output_files"]
    if isinstance(def_out, type(None)):
        def_out = file_mapping[0]["output_files"]
    return def_out


def add_missing_file_map_keys(file_map: OutMapping) -> OutMapping:
    for map_key in ["input_files", "parameter"]:
        if map_key not in file_map.keys():
            file_map[map_key] = None                         # type:ignore
    return file_map

def autocomplete_file_mapping(file_mapping: List[OutMapping]) -> List[OutMapping]:
    return [add_missing_file_map_keys(file_map) for file_map in file_mapping]



""" All test around the parameter checks"""

from omnibenchmark.management import parameter_checks
import pytest
import re
from omnibenchmark.utils.exceptions import ParameterError


# Test get_all_parameter_combinations
def test_get_all_parameter_combinations_two(mock_param_values):
    param_comb = parameter_checks.get_all_parameter_combinations(mock_param_values)
    assert param_comb[0] == {"param1": 0, "param2": "test"}


def test_get_all_parameter_combinations_one():
    param_comb = parameter_checks.get_all_parameter_combinations(
        values={"param1": [0, 10]}
    )
    assert param_comb == [{"param1": 0}, {"param1": 10}]


def test_get_all_parameter_combinations_multiple(mock_param_values):
    mock_param_values["param3"] = [30, "new_test"]
    param_comb = parameter_checks.get_all_parameter_combinations(mock_param_values)
    assert param_comb[0] == {"param1": 0, "param2": "test", "param3": 30}


def test_get_all_parameter_combinations_none():
    param_comb = parameter_checks.get_all_parameter_combinations(values=None)
    assert param_comb == []


# Test apply_filter
def test_apply_filter_num_bounds():
    value_list = [3, 7, "10", "20", 40, "70"]
    filt = {"lower": 10, "upper": "50", "exclude": 20}
    param_val = parameter_checks.apply_filter(value_list=value_list, filter_vals=filt)
    assert param_val == [10, 40]


def test_apply_filter_filter_list():
    value_list = [3, 7, "10", "20", 40, "70", "test", "another"]
    filt = ["40", 7, "test"]
    param_val = parameter_checks.apply_filter(value_list=value_list, filter_vals=filt)
    assert param_val == [3, "10", "20", "70", "another"]


def test_apply_filter_filter_non_num_bound():
    value_list = [3, "10", 40, "70", "test", "another"]
    filt = {"lower": 10, "upper": "50", "exclude": 20}
    with pytest.raises(ParameterError, match=r"Could not apply filtering of 50*?"):
        parameter_checks.apply_filter(value_list=value_list, filter_vals=filt)


def test_apply_filter_invalid_keys(capsys):
    value_list = [3, 7, "10", "20", 40, "70"]
    filt = {"lower": 10, "upper": "50", "not": 20}
    param_val = parameter_checks.apply_filter(value_list=value_list, filter_vals=filt)
    captured = capsys.readouterr()
    assert param_val == [10, 20, 40]
    assert re.match(r"WARNING: Invalid filter keys.\n*?", captured.out)


# Test filter_parameter
def test_filter_parameter_works(mock_param_values):

    filt = {"param1": {"lower": 5}, "param2": "test"}
    param_val = parameter_checks.filter_parameter(mock_param_values, filter=filt)
    assert param_val == {"param1": [10], "param2": []}


def test_filter_parameter_non_ex_param(mock_param_values, capsys):

    filt = {"param1": {"lower": 5}, "param3": "test"}
    param_val = parameter_checks.filter_parameter(mock_param_values, filter=filt)
    captured = capsys.readouterr()
    assert param_val == {"param1": [10], "param2": ["test"]}
    assert re.match(r"WARNING: Could not find param3*?", captured.out)


def test_filter_parameter_no_filter(mock_param_values):
    param_val = parameter_checks.filter_parameter(mock_param_values, filter=None)
    assert param_val == mock_param_values


# Test filter_parameter_combinations
def test_filter_parameter_combinations_works(mock_combinations):

    filt = {"file": "omnibenchmark/test/managment/ex_filter.json", "upper": "50"}
    param_comb = parameter_checks.filter_parameter_combinations(
        mock_combinations, filter=filt
    )
    assert param_comb == [{"param1": 0, "param2": "test"}]


def test_filter_parameter_combinations_different_formats(mock_combinations):
    mock_combinations[1]["param1"] = "10"
    filt = {"file": "omnibenchmark/test/managment/ex_filter.json", "upper": "50"}
    param_comb = parameter_checks.filter_parameter_combinations(
        mock_combinations, filter=filt
    )
    assert param_comb == [{"param1": 0, "param2": "test"}]



"""Functions to facilitate automatic input generation from file/object, usually config.yaml"""

from typing import Dict, Mapping, List
from omnibenchmark.utils.exceptions import ParameterError
from renku.ui.api.models.dataset import Dataset
from difflib import SequenceMatcher
import re
import os
import json

def match_files_by_name(file_type_dict: Mapping[str, List[str]]) -> Mapping[str, Mapping]:
    match_dict: Dict = {}
    fi_types = list(file_type_dict.keys())
    fi_start = fi_types[0]
    fil_types = [fi_type for fi_type in fi_types if not fi_type == fi_start]
    for fi_idx, fi in enumerate(file_type_dict[fi_start]):
        group_nam = "inst" + str(fi_idx)
        match_dict[group_nam] = {fi_start: fi}
        for fi_type in fil_types:
            fi_top = ""
            seq_top = 0.0
            for fi_path in file_type_dict[fi_type]:
                fi_name = os.path.basename(fi_path)
                seq_match = SequenceMatcher(None, os.path.basename(fi), fi_name).ratio()
                if seq_match > seq_top:
                    seq_top = seq_match
                    fi_top = fi_path
            match_dict[group_nam][fi_type] = fi_top
    return(match_dict)


def get_input_files_from_prefix(
    input_prefix: Mapping[str, List[str]], keyword: List[str]
) -> Mapping[str, Mapping]:
    input_files: Dict = {}
    datasets = Dataset.list()
    key_data = [
        dataset
        for dataset in datasets
        if any(key in keyword for key in dataset.keywords)
    ]
    for data in key_data:
        tmp_dict: Dict = {}
        input_files[data.name] = {}
        for file_type, prefixes in input_prefix.items():
            prefix_list = (
                [prefixes] if not isinstance(prefixes, list) else prefixes        #type: ignore
            )  
            pat_list = [re.compile(pattern) for pattern in prefix_list]
            in_file = [
                fi.path
                for fi in data.files
                if any(
                    pattern.search(os.path.basename(fi.path)) for pattern in pat_list
                )
            ]
            if len(in_file) > 1:
                tmp_dict[file_type] = in_file
            elif len(in_file) < 1:
                print(
                    f"WARNING:Could not find any input file matching the following pattern {prefixes}.\n"
                    f"Please make sure you specified a correct prefix! \n"
                    f"Input dataset {data.name} will be ignored!"
                )
                break
            else:
                input_files[data.name][file_type] = in_file[0]
        if len(tmp_dict) > 0:
            tmp_dict.update(input_files[data.name])
            group_dict = match_files_by_name(tmp_dict)
            print(
                    f"WARNING: Ambigous input files. Found {in_file} for prefix {prefixes}.\n"
                    f"Automatic group detection for files in {data.name}.\n"
                    f"Please check matches to ensure correct groups: {group_dict}"
                )
            del input_files[data.name]
            input_files.update(group_dict)
    file_types = input_prefix.keys()
    incomplete_data = [
        data
        for data in input_files.keys()
        if not all(fi_type in input_files[data].keys() for fi_type in file_types)
    ]
    for data in incomplete_data:
        del input_files[data]

    print(f"The following datasets are specified as inputs: {input_files.keys()}")
    return input_files


def get_parameter_from_dataset(
    names: List[str], keyword: List[str]
) -> Mapping[str, List]:
    values: Dict = {}
    datasets = Dataset.list()
    key_data = [
        dataset
        for dataset in datasets
        if any(key in keyword for key in dataset.keywords)
    ]
    for data in key_data:
        if not len(data.files) == 1:
            raise ParameterError(
                f"Could not identify parameter json file."
                f"Please check the specified keyword {keyword} and dataset {data}."
            )
        param_file = data.files[0]
        with open(param_file.path) as f:
            param_json = json.load(f)

        for nam in names:
            if nam not in param_json.keys():
                print(
                    f"WARNING: Could not find a matching parameter for {nam}.\n"
                    f"The following parameter exist in {data}: {param_json.keys()} \n"
                    f"Please post an issue to request missing parameter."
                )
                continue
            values[nam] = param_json[nam]

    print(f"The following parameter are used: {values.keys()}")
    return values



"""Renku workflow commands"""

from __future__ import annotations
from typing import List, Optional, Mapping, Union
from omnibenchmark.core.input_classes import PathLike
from renku.command.run import run_command
from renku.command.workflow import execute_workflow_command
from renku.command.update import update_command
from renku.command.command_builder.command import CommandResult
from renku.core import errors
from renku.command.format.activity import tabulate_activities
from omnibenchmark.management.general_checks import is_renku_project
import logging

logger = logging.getLogger("omnibenchmark.renku_commands")


def renku_workflow_run(
    command_line: str,
    name: str = None,
    description: str = None,
    keyword: str = None,
    success_codes: List[int] = None,
    explicit_inputs: List = [],
    explicit_outputs: List = [],
    explicit_parameters: List = [],
    no_output: bool = False,
    no_input_detection: bool = False,
    no_output_detection: bool = False,
) -> CommandResult:

    if not is_renku_project():
        logger.error(
            f"Project is not a renku project. No workflow generated, consider to running your command without renku."
        )
        return

    workflow = (
        run_command()
        .build()
        .execute(
            name=name,
            description=description,
            keyword=keyword,
            explicit_inputs=explicit_inputs,
            explicit_outputs=explicit_outputs,
            explicit_parameters=explicit_parameters,
            no_output=no_output,
            no_input_detection=no_input_detection,
            no_output_detection=no_output_detection,
            success_codes=success_codes,
            command_line=command_line,
        )
    )

    return workflow


def renku_workflow_execute(
    name_or_id: str,
    set_params: List[str],
    provider: str = "cwltool",
    config: Optional[str] = None,
    values: Optional[str] = None,
):
    workflow = (
        execute_workflow_command()
        .build()
        .execute(
            name_or_id=name_or_id,
            set_params=set_params,
            provider=provider,
            config=config,
            values=values,
        )
    )
    return workflow


def renku_update_activity(
    update_all: bool = False,
    paths: Optional[List[PathLike]] = None,
    dry_run: bool = False,
    provider: str = "cwltool",
    ignore_deleted: bool = True,
    config: Optional[str] = None,
):
    try:
        result = (
            update_command()
            .build()
            .execute(
                update_all=update_all,
                dry_run=dry_run,
                paths=paths,
                provider=provider,
                config=config,
                ignore_deleted=ignore_deleted,
            )
        )
    except:
        print(f"Nothing to update/execute for {paths}.")
    else:
        if dry_run:
            activities, modified_inputs = result.output
            print(tabulate_activities(activities, modified_inputs))



"""General context manager used in omnibenchmark"""

from renku.ui.api.models.project import Project
from renku.command.command_builder.database_dispatcher import DatabaseDispatcher
from renku.infrastructure.gateway.activity_gateway import ActivityGateway
from renku.infrastructure.gateway.plan_gateway import PlanGateway


def get_database_dispatcher() -> DatabaseDispatcher:
    """Configure database from project

    Returns:
        DatabaseDispatcher: A renku database dispatcher to find project metadata.
    """
    # Get project context
    project = Project()
    client = project.client
    database_dispatcher = DatabaseDispatcher()
    database_dispatcher.push_database_to_stack(client.database_path)
    return database_dispatcher


def configure_activity_gateway() -> ActivityGateway:
    """Configure activity gateway from project

    Returns:
        ActivityGateway: A renku ActivityGateway with the currents project as underlying database.
    """
    # Get project context
    database_dispatcher = get_database_dispatcher()
    activity_gateway = ActivityGateway()
    activity_gateway.database_dispatcher = database_dispatcher
    return activity_gateway


def configure_plan_gateway() -> PlanGateway:
    """Configure plan gateway from project

    Returns:
        PlanGateway: A renku PlanGateway with the currents project as underlying database.
    """
    # Get project context
    database_dispatcher = get_database_dispatcher()
    plan_gateway = PlanGateway()
    plan_gateway.database_dispatcher = database_dispatcher
    return plan_gateway



"""Everything related to parameter filtering and handling"""

from typing import Any, Mapping, List, Optional, Dict
from omnibenchmark.utils.exceptions import ParameterError
import itertools
import json


def apply_filter(value_list: List, filter_vals) -> List:
    if isinstance(filter_vals, Dict):
        if any(key not in ["upper", "lower", "exclude"] for key in filter_vals.keys()):
            print(
                f"WARNING: Invalid filter keys.\n"
                f"Please use 'upper', 'lower', 'exclude' only: {filter_vals.keys()}"
            )
        if "upper" in filter_vals.keys():
            try:
                value_list = [
                    int(x) for x in value_list if int(x) <= int(filter_vals["upper"])
                ]
            except:
                raise ParameterError(
                    f'Could not apply filtering of {filter_vals["upper"]}.\n'
                    f"Please check if all values can be converted into a numeric: {value_list}.\n"
                    f'Use the "except" filer key otherwise.'
                )

        if "lower" in filter_vals.keys():
            try:
                value_list = [
                    int(x) for x in value_list if int(x) >= int(filter_vals["lower"])
                ]
            except:
                raise ParameterError(
                    f'Could not apply filtering of {filter_vals["lower"]}.\n'
                    f"Please check if all values can be converted into a numeric: {value_list}.\n"
                    f'Use the "except" filer key otherwise.'
                )
        if "exclude" in filter_vals.keys():
            if isinstance(filter_vals["exclude"], List):
                filter_ex_string = [str(x) for x in filter_vals["exclude"]]
            else:
                filter_ex_string = str(filter_vals["exclude"])  # type: ignore
            value_list = [x for x in value_list if str(x) not in filter_ex_string]
    else:
        if isinstance(filter_vals, List):
            filter_list = [str(x) for x in filter_vals]
        else:
            filter_list = str(filter_vals)  # type: ignore
        value_list = [val for val in value_list if str(val) not in filter_list]
    return value_list


def filter_parameter(
    values: Mapping[str, List], filter: Optional[Mapping[str, Any]]
) -> Mapping[str, List]:

    if filter is None:
        return values

    filter_params = [param for param in filter.keys() if not param == "file"]
    for filt in filter_params:
        if filt not in values.keys():
            print(
                f"WARNING: Could not find {filt} in the specified parameter space: {values.keys()}.\n"
                f"No filter is going to be applied on {filt}."
            )
            continue
        values[filt] = apply_filter(values[filt], filter[filt])  # type: ignore
    return values


def get_all_parameter_combinations(
    values: Mapping[str, List]
) -> List[Mapping[str, str]]:
    comb_list: List = []
    all_combinations = []
    if not isinstance(values, type(None)):
        all_combinations = list(itertools.product(*values.values()))
    for comb in all_combinations:
        comb_dict = dict(zip(list(values.keys()), comb))
        comb_list.append(comb_dict)
    return comb_list


def dict_values_to_str(convert_dict: Mapping) -> Mapping:
    return {str(key): str(value) for key, value in convert_dict.items()}


def filter_parameter_combinations(  # type: ignore
    combinations: List[Mapping[str, str]], filter: Optional[Mapping[str, Any]]
) -> List[Mapping[str, str]]:

    if filter is None:
        return combinations
    if "file" not in filter.keys():
        return combinations
    with open(filter["file"]) as f:
        filter_json = json.load(f)
    filter_list = [
        dict_values_to_str(filt) for filt in filter_json["filter_combinations"]
    ]

    return [
        comb for comb in combinations if dict_values_to_str(comb) not in filter_list
    ]



"""Checks related to workflow management"""

from os import PathLike
from typing import List, Optional, Union
from omnibenchmark.utils.context_manager import (
    configure_plan_gateway,
    configure_activity_gateway,
)
from renku.infrastructure.gateway.activity_gateway import ActivityGateway
from renku.domain_model.workflow.plan import Plan
from omnibenchmark.utils.exceptions import WorkflowError
from omnibenchmark.management import general_checks
import logging

logger = logging.getLogger("omnibenchmark.management.wflow_checks")

## Activities


def activity_plan_is_valid(
    output: Union[PathLike, str], activity_gateway: ActivityGateway
) -> Optional[Union[PathLike, str]]:
    """Check if there is a valid activity associated to a specifc output path

    Args:
        output (Path): Path to an output file
        activity_gateway(ActivityGateway): Activity gateway with a configured database_dispatcher (See get_database_dispatcher).

    Returns:
        output, if a valid acitivity exists, else None
    """
    all_activities = (
        activity_gateway.get_activities_by_generation(output)
        if general_checks.is_renku_project()
        else []
    )
    valid_activities = [
        act
        for act in all_activities
        if isinstance(act.association.plan.invalidated_at, type(None))
    ]
    return output if len(valid_activities) >= 1 else None


def filter_activity_exist(outputs: List[str]) -> List[Union[PathLike, str, None]]:
    """Checks a list of outputs for existing and valid acitivites

    Args:
        outputs (List[str]): A List of output paths

    Returns:
        List[str]: A list of output paths with valid activities
    """

    activity_gateway = configure_activity_gateway()
    return [
        activity_plan_is_valid(out, activity_gateway=activity_gateway)
        for out in outputs
    ]


## Plans


def get_all_plans() -> List[Plan]:
    """Get all existing plans in a project

    Returns:
        List[Plan]: List with all existing plans
    """
    if not general_checks.is_renku_project():
        return []
    else:
        plan_gateway = configure_plan_gateway()
        return plan_gateway.get_all_plans()


def filter_valid_plans(plans: List[Plan]) -> List[Plan]:
    """Filter a list of plans to contain only those plans that were not invalidated

    Args:
        plans (List[Plan]): A list of renku plans

    Returns:
        List[Plan]: A list of valid renku plans
    """
    return [plan for plan in plans if isinstance(plan.invalidated_at, type(None))]


def get_plan_output_values(plan: Plan) -> List[str]:
    """Get default output values for a renku plan

    Args:
        plan (Plan): A renku plan

    Returns:
        str: The default output value of the renku plan
    """
    return [out_val.default_value for out_val in plan.outputs]


def get_plan_id_from_output(output: str, activity_gateway=ActivityGateway) -> List[str]:
    """Get default output values for a renku plan

    Args:
        plan (Plan): A renku plan

    Returns:
        str: The default output value of the renku plan
    """

    all_activities = (
        activity_gateway.get_activities_by_generation(output)
        if general_checks.is_renku_project()
        else []
    )
    return [act.association.plan.id for act in all_activities]


def find_plan_by_outputs(outputs: List[str]) -> Plan:
    """Find a plan that is related to any paths that are part of a list of output files

    Args:
        outputs (List[str]): List of output files

    Returns:
        Optional[Plan]: Plan to generate one or several of those output files
    """
    plans = get_all_plans()
    valid_plans = filter_valid_plans(plans)
    activity_gateway = configure_activity_gateway()
    out_plan_ids = [
        get_plan_id_from_output(out, activity_gateway=activity_gateway)
        for out in outputs
    ]
    flat_ids = [item for sublist in out_plan_ids for item in sublist]
    return [plan for plan in valid_plans if plan.id in flat_ids]


def check_plan_exist(out_files: List[str]) -> Optional[Plan]:
    """Check if exactly one valid plan exists, that describes generation of one or more of the output files

    Args:
        out_files (List[str]): List of output files to look for associated plans

    Returns:
        Optional[Plan]: If one valid plan can be identified, it will be returned.
    """

    plan = find_plan_by_outputs(out_files)

    if len(plan) > 1:
        raise WorkflowError(
            "Ambiguity: To many plans found associated to output files. Make sure that all old plans have been invalidated or specify the correct plan as input."
        )

    if len(plan) == 1:
        logger.info(f"Plan with id {plan[0].id} and name {plan[0].name} exist")
        return plan[0]
    else:
        return None



from omnibenchmark.core.input_classes import OmniInput, OmniParameter
from omnibenchmark.core.output_classes import OmniCommand, OmniOutput
from omnibenchmark.core.omni_object import OmniObject
import omnibenchmark.utils.build_omni_object as omni
from omnibenchmark.utils.exceptions import InputError
from typeguard import check_type
import pytest
import yaml


# Test build_omni_input_from_config_inputs
def test_build_omni_input_from_config_inputs_with_input(mock_config):
    omni_input = omni.build_omni_input_from_config_inputs(mock_config["inputs"])
    assert isinstance(omni_input, OmniInput)
    assert omni_input.names == ["count_file", "dim_red_file"]
    assert omni_input.prefix == {
        "count_file": ["counts"],
        "dim_red_file": ["features", "genes"],
    }
    assert omni_input.default_files == {
        "count_file": "data/import_this_dataset/import_this_dataset__counts.mtx.gz",
        "dim_red_file": "data/import_this_dataset/import_this_dataset__dim_red_file.json",
    }


def test_build_omni_input_from_config_inputs_with_none():
    assert isinstance(
        omni.build_omni_input_from_config_inputs(config_inputs=None), type(None)
    )


def test_build_omni_input_from_config_inputs_no_keyword(mock_config):
    del mock_config["inputs"]["keywords"]
    omni_input = omni.build_omni_input_from_config_inputs(mock_config["inputs"])
    assert isinstance(omni_input, OmniInput)


def test_build_omni_input_from_config_inputs_no_files(mock_config):
    del mock_config["inputs"]["files"]
    omni_input = omni.build_omni_input_from_config_inputs(mock_config["inputs"])
    assert isinstance(omni_input, OmniInput)
    assert omni_input.names == ["count_file", "dim_red_file"]


def test_build_omni_input_from_config_inputs_no_files_pre(mock_config):
    del mock_config["inputs"]["files"]
    del mock_config["inputs"]["prefix"]
    omni_input = omni.build_omni_input_from_config_inputs(mock_config["inputs"])
    assert isinstance(omni_input, OmniInput)


def test_build_omni_input_from_config_inputs_minimal(mock_config):
    del mock_config["inputs"]["input_files"]
    del mock_config["inputs"]["default"]
    omni_input = omni.build_omni_input_from_config_inputs(mock_config["inputs"])
    assert isinstance(omni_input, OmniInput)
    assert omni_input.names == ["count_file", "dim_red_file"]


def test_build_omni_input_from_config_inputs_none(mock_config):
    del mock_config["inputs"]["input_files"]
    del mock_config["inputs"]["prefix"]
    del mock_config["inputs"]["files"]
    omni_input = omni.build_omni_input_from_config_inputs(mock_config["inputs"])
    assert isinstance(omni_input, type(None))


# Test build_omni_parameter_from_config_params
def test_build_omni_parameter_from_config_params_with_params(mock_config):
    assert isinstance(
        omni.build_omni_parameter_from_config_params(mock_config["parameter"]),
        OmniParameter,
    )


def test_build_omni_parameter_from_config_params_none():
    assert isinstance(
        omni.build_omni_parameter_from_config_params(config_params=None), type(None)
    )


def test_build_omni_parameter_from_config_params_no_values(mock_config):
    del mock_config["parameter"]["values"]
    omni_param = omni.build_omni_parameter_from_config_params(mock_config["parameter"])
    assert isinstance(omni_param, OmniParameter)


def test_build_omni_parameter_from_config_params_no_combinations(mock_config):
    del mock_config["parameter"]["combinations"]
    omni_param = omni.build_omni_parameter_from_config_params(mock_config["parameter"])
    assert isinstance(omni_param, OmniParameter)


def test_build_omni_parameter_from_config_params_no_default(mock_config):
    del mock_config["parameter"]["default"]
    del mock_config["parameter"]["filter"]
    omni_param = omni.build_omni_parameter_from_config_params(mock_config["parameter"])
    assert isinstance(omni_param, OmniParameter)


def test_build_omni_parameter_from_config_params_no_filter(mock_config):
    del mock_config["parameter"]["filter"]
    omni_param = omni.build_omni_parameter_from_config_params(mock_config["parameter"])
    assert isinstance(omni_param, OmniParameter)


def test_build_omni_parameter_from_config_params_no_key(mock_config):
    del mock_config["parameter"]["keywords"]
    omni_param = omni.build_omni_parameter_from_config_params(mock_config["parameter"])
    assert isinstance(omni_param, OmniParameter)


### Test build_omni_parameter_from_config_params get names
def test_build_omni_parameter_from_config_params_no_names_at_all(mock_config):
    del mock_config["parameter"]["names"]
    del mock_config["parameter"]["combinations"]
    del mock_config["parameter"]["default"]
    del mock_config["parameter"]["values"]
    omni_param = omni.build_omni_parameter_from_config_params(mock_config["parameter"])
    assert isinstance(omni_param, type(None))


def test_build_omni_parameter_from_config_params_names_from_combs(mock_config):
    del mock_config["parameter"]["names"]
    del mock_config["parameter"]["default"]
    del mock_config["parameter"]["values"]
    del mock_config["parameter"]["filter"]
    omni_param = omni.build_omni_parameter_from_config_params(mock_config["parameter"])
    assert isinstance(omni_param, OmniParameter)


def test_build_omni_parameter_from_config_params_names_from_values(mock_config):
    del mock_config["parameter"]["names"]
    del mock_config["parameter"]["default"]
    del mock_config["parameter"]["combinations"]
    del mock_config["parameter"]["filter"]
    omni_param = omni.build_omni_parameter_from_config_params(mock_config["parameter"])
    assert isinstance(omni_param, OmniParameter)


def test_build_omni_parameter_from_config_params_names_from_default(mock_config):
    del mock_config["parameter"]["names"]
    del mock_config["parameter"]["values"]
    del mock_config["parameter"]["combinations"]
    omni_param = omni.build_omni_parameter_from_config_params(mock_config["parameter"])
    assert isinstance(omni_param, OmniParameter)


# Test build_omni_output_from_config
def test_build_omni_output_from_config_with_outputs(mock_config):
    omni_output = omni.build_omni_output_from_config(mock_config)
    assert isinstance(omni_output, OmniOutput)
    assert omni_output.file_mapping[0]["parameter"]["param1"] == 10


def test_build_omni_output_from_config_no_in_params(mock_config):
    del mock_config["parameter"]
    del mock_config["inputs"]
    assert isinstance(omni.build_omni_output_from_config(mock_config), OmniOutput)


def test_build_omni_output_from_config_no_outputs(mock_config):
    del mock_config["parameter"]
    del mock_config["inputs"]
    del mock_config["outputs"]
    assert isinstance(omni.build_omni_output_from_config(mock_config), type(None))


def test_build_omni_output_from_config_no_name(mock_config):
    del mock_config["data"]["name"]
    with pytest.raises(InputError, match=r"Invalid config file:*?"):
        omni.build_omni_output_from_config(mock_config)


def test_build_omni_output_from_config_no_mapping(mock_config):
    del mock_config["parameter"]
    del mock_config["inputs"]
    del mock_config["outputs"]["file_mapping"]
    assert isinstance(omni.build_omni_output_from_config(mock_config), OmniOutput)


# Test build_omni_command_from_config
def test_build_omni_command_from_config_works(mock_config):
    assert isinstance(omni.build_omni_command_from_config(mock_config), OmniCommand)


def test_build_omni_command_from_config_no_command_line(mock_config):
    del mock_config["command_line"]
    omni_com = omni.build_omni_command_from_config(mock_config)
    assert isinstance(omni_com, OmniCommand)


def test_build_omni_command_from_config_no_output(mock_config):
    del mock_config["outputs"]
    omni_com = omni.build_omni_command_from_config(mock_config)
    assert isinstance(omni_com, OmniCommand)


def test_build_omni_command_from_config_no_script(mock_config):
    del mock_config["script"]
    omni_com = omni.build_omni_command_from_config(mock_config)
    assert isinstance(omni_com, type(None))


# Test build_omni_object_from_config
def test_build_omni_object_from_config_works(mock_config):
    omni_obj = omni.build_omni_object_from_config(mock_config)
    assert isinstance(omni_obj, OmniObject)
    assert isinstance(omni_obj.inputs, OmniInput)
    assert isinstance(omni_obj.outputs, OmniOutput)
    assert isinstance(omni_obj.parameter, OmniParameter)


def test_build_omni_object_from_config_no_input(mock_config):
    del mock_config["inputs"]
    omni_obj = omni.build_omni_object_from_config(mock_config)
    assert isinstance(omni_obj, OmniObject)
    assert isinstance(omni_obj.inputs, type(None))


def test_build_omni_object_from_config_no_output(mock_config):
    del mock_config["outputs"]
    omni_obj = omni.build_omni_object_from_config(mock_config)
    assert isinstance(omni_obj, OmniObject)
    assert isinstance(omni_obj.outputs, type(None))


def test_build_omni_object_from_config_no_param(mock_config):
    del mock_config["parameter"]
    omni_obj = omni.build_omni_object_from_config(mock_config)
    assert isinstance(omni_obj, OmniObject)
    assert isinstance(omni_obj.parameter, type(None))


def test_build_omni_object_from_config_minimal(mock_config):
    del mock_config["parameter"]
    del mock_config["inputs"]
    del mock_config["outputs"]
    omni_obj = omni.build_omni_object_from_config(mock_config)
    assert isinstance(omni_obj, OmniObject)


@pytest.mark.api_call
def test_build_omni_object_from_config_no_orchestrator(mock_config):
    del mock_config["orchestrator"]
    omni_obj = omni.build_omni_object_from_config(mock_config)
    assert isinstance(omni_obj, OmniObject)
    assert omni_obj.orchestrator is not None


def test_build_omni_object_from_config_no_orchestrator_bench_name(mock_config):
    del mock_config["orchestrator"]
    del mock_config["benchmark_name"]
    omni_obj = omni.build_omni_object_from_config(mock_config)
    assert isinstance(omni_obj, OmniObject)
    assert omni_obj.orchestrator is None


def test_build_omni_object_from_config_no_name_fails(mock_config):
    del mock_config["data"]
    with pytest.raises(InputError, match=r"Invalid config file:*?"):
        omni_obj = omni.build_omni_object_from_config(mock_config)


def test_build_omni_object_from_config_yaml_template_fun():
    def dummy_fun(com_mapping):
        return com_mapping

    yaml_file = "omnibenchmark/test/utils/ex_config.yaml"
    with open(yaml_file) as f:
        config = yaml.load(f, Loader=yaml.FullLoader)
    config["outputs"]["template_fun"] = dummy_fun
    check_type("config", config, omni.ConfigDict)
    omni_obj = omni.build_omni_object_from_config(config)
    assert isinstance(omni_obj, OmniObject)


# Test get_omni_object_from_yaml
def test_get_omni_object_from_yaml_works():
    omni_obj = omni.get_omni_object_from_yaml("omnibenchmark/test/utils/ex_config.yaml")
    assert isinstance(omni_obj, OmniObject)


def test_get_omni_object_from_yaml_no_input(monkeypatch, mock_config):
    def return_mock_config(*args, **kwargs):
        m_conf = mock_config
        del m_conf["inputs"]
        return m_conf

    monkeypatch.setattr(yaml, "load", return_mock_config)

    omni_obj = omni.get_omni_object_from_yaml("omnibenchmark/test/utils/ex_config.yaml")
    assert isinstance(omni_obj, OmniObject)


def test_get_omni_object_from_yaml_random_key(monkeypatch, mock_config):
    def return_mock_config(*args, **kwargs):
        m_conf = mock_config
        m_conf["random"] = "something"
        return m_conf

    monkeypatch.setattr(yaml, "load", return_mock_config)
    with pytest.raises(TypeError, match=r"extra key(s)*?"):
        omni_obj = omni.get_omni_object_from_yaml(
            "omnibenchmark/test/utils/ex_config.yaml"
        )


def test_get_omni_object_from_yaml_type_mismatch(monkeypatch, mock_config):
    def return_mock_config(*args, **kwargs):
        m_conf = mock_config
        m_conf["data"]["name"] = 10
        return m_conf

    monkeypatch.setattr(yaml, "load", return_mock_config)
    with pytest.raises(TypeError, match=r"type of dict item*?"):
        omni_obj = omni.get_omni_object_from_yaml(
            "omnibenchmark/test/utils/ex_config.yaml"
        )



import logging
from omnibenchmark.core.input_classes import OmniInput, OmniParameter
from omnibenchmark.core.output_classes import OmniCommand, OmniOutput, OmniPlan
from omnibenchmark.renku_commands.datasets import (
    renku_dataset_create,
    renku_dataset_update,
)
from omnibenchmark.management.run_commands import (
    manage_renku_plan,
    manage_renku_activities,
    check_omni_command,
    get_all_output_file_names,
    check_output_directories,
)
from omnibenchmark.management.general_checks import find_orchestrator, is_renku_project
from omnibenchmark.management.data_commands import update_dataset_files
from omnibenchmark.utils.auto_output import get_default, get_default_input
from omnibenchmark.utils.user_input_checks import empty_object_to_none
from omnibenchmark.utils.exceptions import InputError
from omnibenchmark.utils.default_global_vars import OmniRenkuInst
from renku.domain_model.dataset import Dataset as RenkuDataSet
from typing import List, Optional
from os import PathLike


class OmniObject(OmniRenkuInst):

    """OmniBench object

    An Omnibenchmark component.
    """

    def __init__(
        self,
        name: str,
        keyword: Optional[List[str]] = None,
        title: Optional[str] = None,
        description: Optional[str] = None,
        script: Optional[PathLike] = None,
        command: Optional[OmniCommand] = None,
        inputs: Optional[OmniInput] = None,
        parameter: Optional[OmniParameter] = None,
        outputs: Optional[OmniOutput] = None,
        omni_plan: Optional[OmniPlan] = None,
        benchmark_name: Optional[str] = None,
        orchestrator: Optional[str] = None,
    ):

        self.logger = logging.getLogger("omnibenchmark.OmniObject")
        self.name = name
        self.keyword = keyword
        self.title = title
        self.description = description
        self.command = command
        self.inputs = inputs
        self.outputs = outputs
        self.parameter = parameter
        self.script = script
        self.omni_plan = omni_plan
        self.orchestrator = orchestrator
        self.benchmark_name = benchmark_name
        self.renku: bool = is_renku_project()
        self.kg_url: str = super().KG_URL

        if self.command is None and self.script is not None:
            self.command = OmniCommand(script=self.script, outputs=self.outputs)

        if self.orchestrator is None and self.benchmark_name is not None:
            self.orchestrator = find_orchestrator(
                benchmark_name=self.benchmark_name, bench_url=self.BENCH_URL
            )

    def create_dataset(self) -> RenkuDataSet:
        """create renku-dataset defined by the attributes of the class instance

        Returns:
            RenkuDataSet: An object of class Dataset from renku.core.models.dataset containing the class instances attributes.
        """
        renku_dataset = renku_dataset_create(
            self.name,
            self.kg_url,
            title=self.title,
            description=self.description,
            keyword=self.keyword,
        )
        return renku_dataset

    def run_renku(self):
        self.command = check_omni_command(self.command, self.script, self.outputs)
        out_files = get_all_output_file_names(self.outputs)
        check_output_directories(out_files)
        input_default = empty_object_to_none(get_default_input(self.inputs))
        param_default = empty_object_to_none(get_default(self.parameter))
        out_default = empty_object_to_none(get_default(self.outputs))
        self.omni_plan = manage_renku_plan(
            out_files=out_files,
            omni_plan=self.omni_plan,
            command=self.command.command_line,
            name=self.name,
            description=self.description,
            default_output=out_default,
            default_input=input_default,
            default_parameter=param_default,
        )
        manage_renku_activities(outputs=self.outputs, omni_plan=self.omni_plan)

    def update_result_dataset(self):
        out_files = get_all_output_file_names(self.outputs)
        update_dataset_files(urls=out_files, dataset_name=self.name)
        renku_dataset_update(names=[self.name])

    def update_object(self):
        if self.orchestrator is None:
            if self.benchmark_name is not None:
                self.orchestrator = find_orchestrator(
                    benchmark_name=self.benchmark_name, bench_url=self.BENCH_URL
                )
            if self.orchestrator is None:
                raise InputError(
                    f"Can not update object without specification of benchmark. \n"
                    f"Please indicate which benchmark this object is associated to by running: \n"
                    f"OmniObject.orchestrator = find_orchestrator(BENCHMARK_NAME) \n"
                    f"Look at {self.BENCH_URL} to get a list of possible BENCHMARK_NAMEs"
                )
        if self.inputs is not None:
            self.inputs.update_inputs(
                orchestrator=self.orchestrator,
                query_url=self.DATA_QUERY_URL,
                data_url=self.DATA_URL,
                gitlab_url=self.GIT_URL,
            )
        if self.parameter is not None:
            self.parameter.update_parameter(
                orchestrator=self.orchestrator,
                query_url=self.DATA_QUERY_URL,
                data_url=self.DATA_URL,
                gitlab_url=self.GIT_URL,
            )
        if self.outputs is not None:
            self.outputs.inputs = self.inputs
            self.outputs.parameter = self.parameter
            self.outputs.update_outputs()

        #return self

    def execute_plan_without_KG(
        self, inputs, parameter, outputs, time_out: Optional[int] = None
    ):
        pass



""" All functions around automatization of runs/activities and updates of those"""
from typing import Iterable, Mapping, List, Optional, Union, Dict
from renku.domain_model.workflow.plan import Plan
from omnibenchmark.core.input_classes import OutMapping
from collections import defaultdict
from renku.core.util.urls import get_slug
from omnibenchmark.utils.user_input_checks import flatten
from os import PathLike


def map_plan_names(plan: Plan) -> Mapping[str, str]:
    plan_map = {}
    for input in plan.inputs:
        plan_map[input.name] = input.default_value
    for output in plan.outputs:
        plan_map[output.name] = output.default_value
    for parameter in plan.parameters:
        plan_map[parameter.name] = parameter.default_value

    return plan_map


def get_file_type_dict(file_mapping: OutMapping) -> Mapping[str, str]:
    type_map: Dict = {}
    for file_dict in list(file_mapping.values()):
        if file_dict is not None:
            for file_type, file_name in file_dict.items():  # type: ignore
                type_map[str(file_type)] = file_name
    return type_map


def get_file_name_dict(
    file_mapping: List[Mapping[str, Optional[str]]]
) -> Mapping[str, str]:
    map_dict: Dict = defaultdict(list)
    for mapping in file_mapping:
        joined_dict = {
            str(file_name): file_type for file_type, file_name in mapping.items()
        }
        map_dict.update(joined_dict)
    return map_dict


def map_plan_names_by_prefix(plan: Plan, plan_dict=Mapping[str, str]) -> Dict[str, str]:
    map_dict: Dict = {}
    rev_dict: Dict = {}
    for key, value in plan_dict.items():
        rev_dict.setdefault(value, set()).add(key)

    non_unique = flatten([val for val in rev_dict.values() if len(val) > 1])
    plan_elements = [
        item
        for sublist in [plan.inputs, plan.outputs, plan.parameters]
        for item in sublist
    ]

    for plan_name in non_unique:
        plan_element = [
            plan_ele for plan_ele in plan_elements if plan_ele.name == plan_name
        ]
        if len(plan_element) == 1 and not isinstance(
            plan_element[0].prefix, type(None)
        ):
            file_type = get_slug(
                plan_element[0].prefix.strip(" -="), invalid_chars=["."]
            )
            map_dict[file_type] = plan_name
    return map_dict


def map_plan_names_file_types(
    plan: Plan, file_dict: Mapping[str, str]
) -> Mapping[str, str]:
    plan_dict = map_plan_names(plan)
    map_dict = map_plan_names_by_prefix(plan, plan_dict)
    for plan_name, plan_element in plan_dict.items():
        file_type = file_dict[str(plan_element)]
        if len(file_type) > 0 and file_type not in map_dict:
            map_dict[file_type] = plan_name
    return map_dict


def get_file_mapping_from_out_files(
    out_files: List[Union[PathLike, str, None]], file_mapping=Optional[List[OutMapping]]
) -> List[OutMapping]:
    try:
        map_list = [
            mapping
            for mapping in file_mapping
            if any(
                out_file in out_files
                for out_file in list(mapping["output_files"].values())
            )
        ]
    except:
        map_list = []
    finally:
        return map_list


def get_file_list_from_out_mapping(
    out_mapping=Optional[List[OutMapping]],
) -> Optional[List[str]]:
    all_files = []
    for mapping in out_mapping:
        try:
            file_list: Iterable = [
                file_path
                for file_dict in mapping.values()
                for file_path in file_dict.values()
            ]
        except:
            file_list: Iterable = []  # type:ignore
        finally:
            all_files.append(file_list)
    return flatten(all_files)



""" All kind of check or test functions to ensure correct variable inputs beyont type hinting"""

from typing import List, Iterable, Mapping, Union, Optional, overload, Any
from omnibenchmark.utils.exceptions import InputError


def flatten(nested_list: List[Iterable]) -> List:
    """Flatten a nested list into a single list

    Args:
        nested_list (List[List]): A nested list

    Returns:
        List: A flat list
    """
    return [item for sublist in nested_list for item in sublist]


def check_name_matching(entry1: Iterable[str], entry2: Iterable[str]):
    """Check if all entries of a list (entry2) are part of a second list (entry1) of entries

    Args:
        entry1 (List[str]): List to with variables to check
        entry2 (List[str]): List containing the variables that are allowed

    """
    if any(val not in entry1 for val in entry2):
        uniq_entry2 = list(set(entry2))
        raise InputError(f"Expected inputs from ${entry1}, but found ${uniq_entry2}")
    else:
        pass


@overload
def empty_object_to_none(obj: List) -> Optional[List]:
    ...


@overload
def empty_object_to_none(obj: Mapping) -> Optional[Mapping]:
    ...


def empty_object_to_none(obj: Union[List, Mapping]) -> Optional[Union[List, Mapping]]:
    if len(obj) == 0:
        return None
    else:
        return obj


def parse_explicit_inputs(obj: Optional[Mapping]) -> List:
    if obj is None:
        return []
    else:
        return list(obj.values())



"""Defintions and utils concering exceptions"""

import textwrap


class InterpreterError(Exception):
    @staticmethod
    def format_arg(arg):
        if isinstance(arg, str):
            return arg
        elif isinstance(arg, InterpreterError):
            spec = ""
            if arg.interpreter is not None:
                spec += "Interpreter {}".format(arg.interpreter)
            if arg.extension is not None:
                spec += "Extension {}".format(arg.extension)
            if arg.command is not None:
                spec += "Command {}".format(arg.command)

            if spec:
                spec = " ({})".format(spec)

            return "{}{}:\n{}".format(
                arg.__class__.__name__, spec, textwrap.indent(str(arg), "    ")
            )
        else:
            return "{}: {}".format(arg.__class__.__name__, str(arg))

    def __init__(self, *args, interpreter=None, extension=None, command=None):
        super().__init__("\n".join(self.format_arg(arg) for arg in args))

        self.interpreter = interpreter
        self.extension = extension
        self.command = command


class WorkflowError(Exception):
    pass


class InputError(Exception):
    pass


class OutputError(Exception):
    pass


class ProjectError(Exception):
    pass


class ParameterError(Exception):
    pass



from typing import List
import omnibenchmark.utils.auto_output as omni
import pytest

### Test auto output functions

# join_parameter
def test_join_parameter_with_parameter(mock_param):
    test_join = omni.join_parameter(mock_param)
    assert test_join == "param_str_value_str__param_num_10"


def test_join_parameter_without_parameter():
    test_join = omni.join_parameter()
    assert test_join == ""


# join_inputs_parameter
def test_join_inputs_parameter_with_both(mock_param):
    test_join = omni.join_inputs_parameter("test_in", mock_param)
    assert test_join == "test_in__param_str_value_str__param_num_10"


def test_join_inputs_parameter_with_parameter(mock_param):
    test_join = omni.join_inputs_parameter(parameter=mock_param)
    assert test_join == "param_str_value_str__param_num_10"


def test_join_inputs_parameter_with_input():
    test_join = omni.join_inputs_parameter(input="test_in")
    assert test_join == "test_in"


def test_join_inputs_parameter_with_none():
    test_join = omni.join_inputs_parameter()
    assert test_join == ""


# get_out_names_from_input_params
def test_get_out_names_from_input_params_name_out_end(mock_out_end):
    test_out_name = omni.get_out_names_from_input_params(
        name="test_in", output_end=mock_out_end
    )
    assert test_out_name == {
        "red_dim": "data/test_in/test_in__red_dim.mtx.gz",
        "counts": "data/test_in/test_in__counts.json",
    }


def test_get_out_names_from_input_params_remove_dubble_points(mock_out_end):
    mock_out_end["red_dim"] = ".mtx.gz"
    test_out_name = omni.get_out_names_from_input_params(
        name="test_in", output_end=mock_out_end
    )
    assert test_out_name == {
        "red_dim": "data/test_in/test_in__red_dim.mtx.gz",
        "counts": "data/test_in/test_in__counts.json",
    }


def test_get_out_names_from_input_params_name_out_end_parameter(
    mock_out_end, mock_param
):
    test_out_name = omni.get_out_names_from_input_params(
        name="test_in", output_end=mock_out_end, parameter=mock_param
    )
    assert test_out_name == {
        "red_dim": "data/test_in/test_in_param_str_value_str__param_num_10_red_dim.mtx.gz",
        "counts": "data/test_in/test_in_param_str_value_str__param_num_10_counts.json",
    }


def test_get_out_names_from_input_params_new_template(mock_out_end):
    test_out_name = omni.get_out_names_from_input_params(
        name="test_in",
        output_end=mock_out_end,
        out_template="${name}_${out_name}.${out_end}",
    )
    assert test_out_name == {
        "red_dim": "test_in_red_dim.mtx.gz",
        "counts": "test_in_counts.json",
    }


def test_get_out_names_from_input_params_new_tem_vars(mock_out_end):
    test_out_name = omni.get_out_names_from_input_params(
        name="test_in",
        output_end=mock_out_end,
        out_template="${name}_${out_name}_${test_name}.${out_end}",
        test_name="new",
    )
    assert test_out_name == {
        "red_dim": "test_in_red_dim_new.mtx.gz",
        "counts": "test_in_counts_new.json",
    }


def test_get_out_names_from_input_params_invalid_template(mock_out_end):
    with pytest.raises(NameError, match=r"Invalid output filename template: .*?"):
        omni.get_out_names_from_input_params(
            name="test_in",
            output_end=mock_out_end,
            out_template="${quatsch}_${out_name}.${out_end}",
        )


# get_input_file_list
def test_get_input_file_list_with_input(mock_omni_input):
    test_input = omni.get_input_file_list(mock_omni_input)
    assert test_input == ["data1", "data2"]


def test_get_input_file_list_no_input():
    test_input = omni.get_input_file_list()
    assert test_input == []


def test_get_input_file_list_with_none_inputfiles(mock_omni_input):
    mock_omni_input.input_files = None
    test_input = omni.get_input_file_list(mock_omni_input)
    assert test_input == []


# get_parameter_combinations
def test_get_parameter_combinations_with_parameter(
    mock_omni_parameter, mock_combinations
):
    test_param = omni.get_parameter_combinations(parameter=mock_omni_parameter)
    assert test_param == mock_combinations


def test_get_parameter_combinations_no_parameter():
    test_param = omni.get_parameter_combinations()
    assert test_param == []


# get_input_parameter_combinations
def test_get_input_parameter_combinations_none():
    comb = omni.get_input_parameter_combinations()
    assert comb == []


def test_get_input_parameter_combinations_both(mock_omni_parameter, mock_omni_input):
    mock_omni_input.input_files = {
        "data1": {"dim_red_file": "path/to/dim/red", "count_file": "path/to/count"}
    }
    comb = omni.get_input_parameter_combinations(
        parameter=mock_omni_parameter, inputs=mock_omni_input
    )
    assert comb == [
        {
            "input_files": {
                "data1": {
                    "dim_red_file": "path/to/dim/red",
                    "count_file": "path/to/count",
                }
            },
            "parameter": {"param1": 0, "param2": "test"},
            "output_files": None,
        },
        {
            "input_files": {
                "data1": {
                    "dim_red_file": "path/to/dim/red",
                    "count_file": "path/to/count",
                }
            },
            "parameter": {"param1": 10, "param2": "test"},
            "output_files": None,
        },
    ]


def test_get_input_parameter_combinations_parameter(mock_omni_parameter):
    comb = omni.get_input_parameter_combinations(parameter=mock_omni_parameter)
    assert comb == [
        {
            "input_files": None,
            "parameter": {"param1": 0, "param2": "test"},
            "output_files": None,
        },
        {
            "input_files": None,
            "parameter": {"param1": 10, "param2": "test"},
            "output_files": None,
        },
    ]


def test_get_input_parameter_combinations_input(mock_omni_input):
    comb = omni.get_input_parameter_combinations(inputs=mock_omni_input)
    assert comb == [
        {
            "input_files": {
                "data1": {
                    "dim_red_file": "path/to/dim/red",
                    "count_file": "path/to/count",
                }
            },
            "parameter": None,
            "output_files": None,
        },
        {
            "input_files": {
                "data2": {
                    "dim_red_file": "path/to/dim/red2",
                    "count_file": "path/to/count2",
                }
            },
            "parameter": None,
            "output_files": None,
        },
    ]


# get_all_output_combinations
def test_get_all_output_combinations_no_in_param():
    all_out = omni.get_all_output_combinations(
        name="test_out", output_end={"dim_red": "mtx.gz", "counts": "json"}
    )
    assert all_out == [
        {
            "input_files": None,
            "output_files": {
                "dim_red": "data/test_out/test_out__dim_red.mtx.gz",
                "counts": "data/test_out/test_out__counts.json",
            },
            "parameter": None,
        }
    ]


def test_get_all_output_combinations_no_in(mock_omni_parameter):
    all_out = omni.get_all_output_combinations(
        name="test_out", output_end={"dim_red": "mtx.gz"}, parameter=mock_omni_parameter
    )
    assert all_out[0] == {
        "input_files": None,
        "output_files": {
            "dim_red": "data/test_out/test_out_param1_0__param2_test_dim_red.mtx.gz"
        },
        "parameter": {"param1": 0, "param2": "test"},
    }


def test_get_all_output_combinations_no_param(mock_omni_input):
    all_out = omni.get_all_output_combinations(
        name="test_out", output_end={"dim_red": "mtx.gz"}, inputs=mock_omni_input
    )
    assert all_out[0] == {
        "input_files": {
            "count_file": "path/to/count",
            "dim_red_file": "path/to/dim/red",
        },
        "output_files": {"dim_red": "data/test_out/test_out_data1_dim_red.mtx.gz"},
        "parameter": None,
    }


def test_get_all_output_combinations_new_template(mock_omni_input):
    all_out = omni.get_all_output_combinations(
        name="test_out",
        output_end={"dim_red": "mtx.gz"},
        inputs=mock_omni_input,
        out_template="data/${name}/${name}_${unique_values}_${out_name}_${test_name}.${out_end}",
        test_name="new",
    )
    assert all_out[0] == {
        "input_files": {
            "count_file": "path/to/count",
            "dim_red_file": "path/to/dim/red",
        },
        "output_files": {"dim_red": "data/test_out/test_out_data1_dim_red_new.mtx.gz"},
        "parameter": None,
    }


def test_get_all_output_combinations_template_function(mock_omni_input):
    def temp_func(comb):
        temp_val = {}
        if len(list(comb["input_files"].values())[0]) > 1:
            temp_val["in_type"] = "multi"
        else:
            temp_val["in_type"] = "single"
        return temp_val

    all_out = omni.get_all_output_combinations(
        name="test_out",
        output_end={"dim_red": "mtx.gz"},
        inputs=mock_omni_input,
        out_template="data/${name}/${name}_${unique_values}_${out_name}_${in_type}.${out_end}",
        template_fun=temp_func,
    )
    assert all_out[0] == {
        "input_files": {
            "count_file": "path/to/count",
            "dim_red_file": "path/to/dim/red",
        },
        "output_files": {
            "dim_red": "data/test_out/test_out_data1_dim_red_multi.mtx.gz"
        },
        "parameter": None,
    }


def test_get_all_output_combinations_template_fun_vars(
    mock_omni_input, mock_omni_parameter
):
    def temp_func(comb, name_dict):
        data_name = list(comb["input_files"].keys())[0]
        if data_name in name_dict.keys():
            data_name = name_dict[data_name]
        temp_val = {
            "data_name": data_name,
            "param1": comb["parameter"]["param1"],
            "param2": comb["parameter"]["param2"],
        }
        return temp_val

    name_dict = {"data1": "positive_control"}

    all_out = omni.get_all_output_combinations(
        name="test_out",
        output_end={"dim_red": "mtx.gz"},
        inputs=mock_omni_input,
        parameter=mock_omni_parameter,
        out_template="data/${name}/${data_name}_${name}_${param1}_${param2}_${out_name}.${out_end}",
        template_fun=temp_func,
        name_dict=name_dict,
    )
    assert all_out[0]["output_files"] == {
        "dim_red": "data/test_out/positive_control_test_out_0_test_dim_red.mtx.gz"
    }
    assert all_out[2]["output_files"] == {
        "dim_red": "data/test_out/data2_test_out_0_test_dim_red.mtx.gz"
    }


# get_default
def test_get_default_input(mock_omni_input):
    assert omni.get_default(mock_omni_input) == "data1"


def test_get_default_parameter(mock_omni_parameter):
    assert omni.get_default(mock_omni_parameter) == {"param1": "0", "param2": "test"}


def test_get_default_none():
    assert omni.get_default() == []


# get_default_outputs
def test_get_default_outputs_file_mapping(mock_out_mapping):
    default_out = omni.get_default_outputs([mock_out_mapping])
    assert default_out == {"out_file1": "path/to/out1", "out_file2": "path/to/out2"}


def test_get_default_outputs_file_mapping_input(mock_out_mapping, mock_omni_input):
    out_map_all = [mock_out_mapping, mock_out_mapping]
    out_map_all[1]["input_files"] = mock_omni_input.default_files
    out_map_all[1]["output_files"] = {"out_file1": "new_path/to/out1"}
    out_map_all[1]["parameter"] = None
    default_out = omni.get_default_outputs(out_map_all, inputs=mock_omni_input)
    assert default_out == {"out_file1": "new_path/to/out1"}


def test_get_default_outputs_file_mapping_input_param(
    mock_out_mapping, mock_omni_input, mock_omni_parameter
):
    out_map_all = [mock_out_mapping, mock_out_mapping]
    out_map_all[1]["input_files"] = mock_omni_input.default_files
    out_map_all[1]["output_files"] = {"out_file1": "new_path/to/out1"}
    out_map_all[1]["parameter"] = mock_omni_parameter.combinations[0]
    default_out = omni.get_default_outputs(
        out_map_all, inputs=mock_omni_input, parameter=mock_omni_parameter
    )
    assert default_out == {"out_file1": "new_path/to/out1"}



"""Commands related to execute a script with renku workflow/run commands"""

from typing import List, Mapping, Optional, Union
from omnibenchmark.core.input_classes import OutMapping
from omnibenchmark.utils.auto_run import (
    get_file_mapping_from_out_files,
    get_file_type_dict,
    get_file_name_dict,
)
from renku.domain_model.workflow.plan import Plan
from omnibenchmark.utils.exceptions import InputError
from omnibenchmark.management import wflow_checks as wflow
import omnibenchmark.renku_commands.workflows as omni_wflow
from omnibenchmark.utils.user_input_checks import parse_explicit_inputs, flatten
from omnibenchmark.core.output_classes import OmniCommand, OmniOutput, OmniPlan
from renku.command.view_model.plan import PlanViewModel
import os
from os import PathLike
import logging

logger = logging.getLogger("omnibenchmark.management.run_commands")


def update_activity(out_map: OutMapping):
    if out_map["output_files"] is not None:
        out_paths = list(out_map["output_files"].values())
    else:
        out_paths = None
    omni_wflow.renku_update_activity(paths=out_paths)


def create_activity(out_map: OutMapping, omni_plan: OmniPlan):
    plan = omni_plan.plan
    map_dict = omni_plan.param_mapping
    file_dict = get_file_type_dict(out_map)
    if map_dict is None or any(
        file_type not in list(map_dict.keys()) for file_type in list(file_dict.keys())
    ):
        raise NameError(
            f"Could not assign all workflow items from: ${omni_plan.param_mapping}."
            f'Please check "renku workflow show ${plan.name}".'
            f"Update OmniObject.omni_plan.param_mapping accordingly."
        )
    params = [
        map_dict[str(file_type)] + "=" + str(file_path)
        for file_type, file_path in file_dict.items()
    ]
    workflow = omni_wflow.renku_workflow_execute(name_or_id=plan.id, set_params=params)
    return workflow


def manage_renku_plan(
    out_files: List[str],
    omni_plan: Optional[OmniPlan],
    command: str,
    name: Optional[str] = None,
    description: Optional[str] = None,
    keyword: Optional[str] = None,
    default_output: Optional[Mapping] = None,
    default_input: Optional[Mapping] = None,
    default_parameter: Optional[Mapping] = None,
) -> Plan:
    """Check if a renku plan for a set of output files exist and generate one from default parameter if not.

    Args:
        out_files (List[str]): All potential output files
        command_line (str): Command line command to generate default plan.
        name (str, optional): Workflow name. Defaults to None.
        description (str, optional): Workflow description. Defaults to None.
        keyword (str, optional): Workflow keyword. Defaults to None.
        default_output (Optional[Mapping[str, Optional[str]]], optional): Default outputs. Defaults to None.
        default_input (Optional[Mapping[str, Optional[str]]], optional): Default inputs. Defaults to None.
        default_parameter (Optional[Mapping[str, Optional[str]]], optional): Default parameter. Defaults to None.

    Returns:
        Plan: A renku plan object.
    """

    plan = wflow.check_plan_exist(out_files)
    if plan is not None:
        plan_view = PlanViewModel.from_plan(plan)
    def_input = parse_explicit_inputs(default_input)
    def_output = parse_explicit_inputs(default_output)
    def_parameter = parse_explicit_inputs(default_parameter)

    if plan is None:
        result = omni_wflow.renku_workflow_run(
            command_line=command,
            name=name,
            description=description,
            keyword=keyword,
            explicit_inputs=def_input,
            explicit_parameters=def_parameter,
            explicit_outputs=def_output,
        )

        plan_view = result.output

    if omni_plan is None or not plan_view == omni_plan.plan:
        omni_plan = OmniPlan(plan=plan_view)
        file_list = [
            file_map
            for file_map in [default_output, default_input, default_parameter]
            if file_map is not None
        ]
        file_dict = get_file_name_dict(file_mapping=file_list)
        omni_plan.param_mapping = omni_plan.predict_mapping_from_file_dict(
            file_dict=file_dict
        )

    return omni_plan


def check_omni_command(
    command: Optional[OmniCommand],
    script: Optional[Union[PathLike, str]],
    outputs=Optional[OmniOutput],
) -> OmniCommand:
    if command is None:
        if script is None:
            raise InputError(
                f"Can not run a command without a specified script. Please specify what script to run."
            )
        command = OmniCommand(script=script, outputs=outputs)
    return command


def get_all_output_file_names(output: OmniOutput) -> List[str]:
    if output.file_mapping is not None:
        return flatten(
            [
                list(fi_mapping["output_files"].values())
                for fi_mapping in output.file_mapping
                if fi_mapping["output_files"] is not None
            ]
        )
    else:
        return []


def manage_renku_activities(outputs: OmniOutput, omni_plan: OmniPlan):
    # Manage activities
    out_files = get_all_output_file_names(outputs)
    activity_out = wflow.filter_activity_exist(out_files)
    activity_map = get_file_mapping_from_out_files(
        out_files=activity_out, file_mapping=outputs.file_mapping
    )
    if outputs.file_mapping is None:
        outputs.file_mapping = []

    no_activities = [out for out in outputs.file_mapping if out not in activity_map]

    for out_map in no_activities:
        create_activity(out_map, omni_plan)

    for activity in activity_map:
        update_activity(activity)


def check_output_directories(out_files: List[str]):
    out_dir_list = [os.path.dirname(out_file) for out_file in out_files]
    out_dirs = set(out_dir_list)
    for dir in out_dirs:
        os.makedirs(dir, exist_ok=True)



"""Fixtures related to core classes"""

from omnibenchmark.core.input_classes import OutMapping
import pytest


@pytest.fixture
def mock_out_mapping():
    map: OutMapping = {
        "output_files": {"out_file1": "path/to/out1", "out_file2": "path/to/out2"},
        "input_files": {"in_file1": "path/to/in1"},
        "parameter": {"param1": "str_value", "param2": 100},
    }

    return map



"""Omnibenchmark"""

# Version of the omnibenchmark package
__version__ = "0.0.6"

from omnibenchmark import core, management, renku_commands, utils

__all__ = ("core", "management", "renku_commands", "utils")



""" Tests related to the command class"""

from omnibenchmark.core.output_classes import OmniCommand

# Test OmniCommand class
def test_omni_command_with_command_line():
    test_command = OmniCommand(
        script="path/to/some/sricpt.jl",
        command_line='julia path/to/some/script.jl --input "in1"',
    )
    assert isinstance(test_command, OmniCommand)


def test_omni_command_with_output_command_line(mock_omni_output):
    test_command = OmniCommand(
        script="path/to/some/sricpt.jl",
        outputs=mock_omni_output,
        command_line='julia path/to/some/script.jl --input "in1"',
    )
    assert test_command.command_line == 'julia path/to/some/script.jl --input "in1"'


def test_omni_command_with_output(mock_omni_output):
    test_command = OmniCommand(
        script="path/to/some/sricpt.jl", outputs=mock_omni_output
    )
    assert (
        test_command.command_line
        == 'julia path/to/some/sricpt.jl --in_file1 "path/to/in1" --out_file1 "path/to/out1" --out_file2 "path/to/out2" --param1 "str_value" --param2 "100"'
    )



import omnibenchmark.utils.auto_input as omni
from omnibenchmark.utils.exceptions import ParameterError
import renku.ui.api.models.dataset
import pytest
import re

### Test auto input functions

# join_parameter
def test_get_input_files_from_prefix_works(mock_api_Dataset, mock_prefix, monkeypatch):
    del mock_prefix["count_file"]

    def get_mock_list():
        return [mock_api_Dataset]

    monkeypatch.setattr(
        renku.ui.api.models.dataset.Dataset,
        "list",
        lambda *args, **kwargs: get_mock_list(),
    )

    test_join = omni.get_input_files_from_prefix(
        input_prefix=mock_prefix, keyword=["mock", "some"]
    )
    assert test_join == {
        "mock_dataset": {"dim_red_file": "some/path/to/genes_file.txt"}
    }


def test_get_input_files_from_prefix_works_incomplete(
    mock_api_Dataset, mock_prefix, monkeypatch, capsys
):
    def get_mock_list():
        return [mock_api_Dataset]

    monkeypatch.setattr(
        renku.ui.api.models.dataset.Dataset,
        "list",
        lambda *args, **kwargs: get_mock_list(),
    )

    test_join = omni.get_input_files_from_prefix(
        input_prefix=mock_prefix, keyword=["mock", "some"]
    )
    captured = capsys.readouterr()
    assert test_join == {}
    assert re.match(
        r"WARNING:Could not find any input file matching the following pattern*?",
        captured.out,
    )


def test_get_input_files_from_prefix_works_regexpr(
    mock_api_Dataset, mock_prefix, monkeypatch
):
    del mock_prefix["count_file"]
    mock_prefix["dim_red_file"] = ["features", "file"]

    def get_mock_list():
        return [mock_api_Dataset]

    monkeypatch.setattr(
        renku.ui.api.models.dataset.Dataset,
        "list",
        lambda *args, **kwargs: get_mock_list(),
    )

    test_join = omni.get_input_files_from_prefix(
        input_prefix=mock_prefix, keyword=["mock", "some"]
    )
    assert test_join == {
        "mock_dataset": {"dim_red_file": "some/path/to/genes_file.txt"}
    }


# get_parameter_from_dataset
def test_get_parameter_from_dataset_works(mock_paramapi_Dataset, monkeypatch):
    mock_paramapi_Dataset._files = [mock_paramapi_Dataset._files[0]]

    def get_mock_list():
        return [mock_paramapi_Dataset]

    monkeypatch.setattr(
        renku.ui.api.models.dataset.Dataset,
        "list",
        lambda *args, **kwargs: get_mock_list(),
    )

    test_join = omni.get_parameter_from_dataset(
        names=["param2"], keyword=["mock_param", "some"]
    )
    assert test_join == {"param2": ["test", "hvg", "dim"]}


def test_get_parameter_from_dataset_to_man_files(mock_paramapi_Dataset, monkeypatch):
    def get_mock_list():
        return [mock_paramapi_Dataset]

    monkeypatch.setattr(
        renku.ui.api.models.dataset.Dataset,
        "list",
        lambda *args, **kwargs: get_mock_list(),
    )
    with pytest.raises(
        ParameterError, match=r"Could not identify parameter json file.*?"
    ):
        omni.get_parameter_from_dataset(
            names=["param2"], keyword=["mock_param", "some"]
        )


def test_get_parameter_from_dataset_not_ex_parameter(
    mock_paramapi_Dataset, monkeypatch, capsys
):
    mock_paramapi_Dataset._files = [mock_paramapi_Dataset._files[0]]

    def get_mock_list():
        return [mock_paramapi_Dataset]

    monkeypatch.setattr(
        renku.ui.api.models.dataset.Dataset,
        "list",
        lambda *args, **kwargs: get_mock_list(),
    )

    test_join = omni.get_parameter_from_dataset(
        names=["param2", "param3"], keyword=["mock_param"]
    )
    captured = capsys.readouterr()
    assert test_join == {"param2": ["test", "hvg", "dim"]}
    assert re.match(
        r"WARNING: Could not find a matching parameter for param3*?", captured.out
    )


def test_get_parameter_from_dataset_not_matching_keyword(
    mock_paramapi_Dataset, monkeypatch
):
    def get_mock_list():
        return [mock_paramapi_Dataset]

    monkeypatch.setattr(
        renku.ui.api.models.dataset.Dataset,
        "list",
        lambda *args, **kwargs: get_mock_list(),
    )

    test_join = omni.get_parameter_from_dataset(
        names=["param2", "param3"], keyword=["quatsch"]
    )
    assert test_join == {}


def test_get_parameter_from_dataset_not_matching_names(
    mock_paramapi_Dataset, monkeypatch
):
    mock_paramapi_Dataset._files = [mock_paramapi_Dataset._files[0]]

    def get_mock_list():
        return [mock_paramapi_Dataset]

    monkeypatch.setattr(
        renku.ui.api.models.dataset.Dataset,
        "list",
        lambda *args, **kwargs: get_mock_list(),
    )

    test_join = omni.get_parameter_from_dataset(
        names=["param4", "param3"], keyword=["mock_param"]
    )
    assert test_join == {}



from omnibenchmark.core.output_classes import OmniCommand, OmniPlan
from omnibenchmark.management import run_commands as omni
import omnibenchmark.renku_commands.workflows as omni_wflow
from omnibenchmark.management import wflow_checks as wflow
from renku.command.command_builder.command import CommandResult
from renku.core.errors import NothingToExecuteError
from omnibenchmark.utils.exceptions import InputError
import pytest


### Test update_activity
@pytest.mark.renku_call
def test_update_activity_no_activity(mock_out_mapping, capsys):
    omni.update_activity(mock_out_mapping)
    captured = capsys.readouterr()
    assert captured.out == r"Nothing to update/execute for*?"


# Test create_activity
def test_create_activity_command(
    mock_out_mapping, mock_plan_dict, mock_plan, monkeypatch
):
    def print_arguments(**kwargs):
        return list(kwargs.values())

    monkeypatch.setattr(omni_wflow, "renku_workflow_execute", print_arguments)
    mock_omni_plan = OmniPlan(plan=mock_plan, param_mapping=mock_plan_dict)
    workflow = omni.create_activity(out_map=mock_out_mapping, omni_plan=mock_omni_plan)

    assert workflow == [
        "XXX",
        [
            "out-1=path/to/out1",
            "out-2=path/to/out2",
            "in-1=path/to/in1",
            "param-1=str_value",
            "param-2=100",
        ],
    ]


def test_create_activity(mock_out_mapping, mock_plan_dict, mock_plan):
    with pytest.raises(NameError, match=r"Could not assign all workflow items from:*?"):
        del mock_plan_dict["in_file1"]
        mock_omni_plan = OmniPlan(plan=mock_plan, param_mapping=mock_plan_dict)
        omni.create_activity(out_map=mock_out_mapping, omni_plan=mock_omni_plan)


# Test manage_renku_plan
def test_manage_renku_plan_with_correct_plan(mock_plan, mock_plan_view, monkeypatch):
    def return_mock_plan(*args, **kwargs):
        return mock_plan

    monkeypatch.setattr(wflow, "check_plan_exist", return_mock_plan)
    mock_omni_plan = OmniPlan(plan=mock_plan_view)
    out_plan = omni.manage_renku_plan(
        out_files=["any", "random"],
        omni_plan=mock_omni_plan,
        command="not_to_run_command_str",
    )
    assert out_plan.plan.id == mock_omni_plan.plan.id


def test_manage_renku_plan_without_any_input(mock_plan_view, monkeypatch):
    def return_run_result(*args, **kwargs):
        return CommandResult(output=mock_plan_view, error=None, status=None)

    monkeypatch.setattr(omni_wflow, "renku_workflow_run", return_run_result)
    mock_omni_plan = OmniPlan(plan=mock_plan_view, param_mapping={})
    out_plan = omni.manage_renku_plan(
        out_files=["any", "random"], omni_plan=None, command="not_to_run_command_str"
    )
    assert out_plan.param_mapping == mock_omni_plan.param_mapping
    assert out_plan.plan == mock_omni_plan.plan


# Test check_omni_command
def test_check_omni_command_with_command():
    mock_command = OmniCommand(
        script="path/to/no/script", command_line="no real command"
    )
    assert (
        omni.check_omni_command(command=mock_command, script=None, outputs=None)
        == mock_command
    )


def test_check_omni_command_no_command_script():
    with pytest.raises(InputError, match=r"Can not run a command*?"):
        assert omni.check_omni_command(command=None, script=None, outputs=None)


# Test get_all_output_file_names
def test_get_all_output_file_names_with_output(mock_omni_output):
    assert omni.get_all_output_file_names(mock_omni_output) == [
        "path/to/out1",
        "path/to/out2",
    ]


def test_get_all_output_file_names_with_no_outfiles(mock_omni_output):
    mock_omni_output.file_mapping = {}
    assert omni.get_all_output_file_names(mock_omni_output) == []


# Test manage_renku_activities
# def test_manage_renku_activities_new_and_existing_activities(mock_omni_output, mock_out_mapping, monkeypatch):
#
#    def return_outfile_list(*args,**kwargs):
#        return ["path/to/out1"]
#
#    monkeypatch.setattr(
#        wflow,
#        "filter_activity_exist",
#        return_outfile_list,
#    )
#    mock_out_mapping2 = mock_out_mapping
#    mock_out_mapping2["output_files"] = {"out_file1": "path/to/out3", "out_file2": "path/to/out4"}
#    mock_omni_output.file_mapping.append(mock_out_mapping)
#
#    def return_arguments(*args, **kwargs):
#        return kwargs.values()
#
#    monkeypatch.setattr(
#        "omnibenchmark.management.run_commands.create_activity",
#        return_arguments,
#    )
#
#    monkeypatch.setattr(
#        "omnibenchmark.management.run_commands.update_activity",
#        return_arguments,
#    )
#    assert omni.manage_renku_activities(mock_omni_output, omni_plan=None) == [mock_out_mapping, mock_out_mapping2]



from omnibenchmark.renku_commands import datasets
import omnibenchmark.management.general_checks
from renku.command.dataset import remove_dataset_command, file_unlink_command
from renku.core import errors
from renku.ui.api import Dataset
from omnibenchmark.utils.exceptions import ProjectError
import pytest
import os
import shutil

### Test renku_dataset_create
def test_renku_dataset_create_for_existing_dataset(get_renkuDataset_List, monkeypatch):
    def get_mock_status():
        return True

    monkeypatch.setattr(
        omnibenchmark.management.general_checks,
        "is_renku_project",
        lambda *args, **kwargs: get_mock_status(),
    )
    return_value = datasets.renku_dataset_create(name="mock_dataset", kg_url="some.url")
    assert isinstance(return_value, type(None))


def test_renku_dataset_create_for_dataset_name_exist(mock_dataset_query, monkeypatch):
    def get_mock_status():
        return True

    monkeypatch.setattr(
        omnibenchmark.management.general_checks,
        "is_renku_project",
        lambda *args, **kwargs: get_mock_status(),
    )

    return_value = datasets.renku_dataset_create(name="mock_dataset", kg_url="some.url")
    assert isinstance(return_value, type(None))


@pytest.mark.renku_call
def test_renku_dataset_create_works():

    data_name = "test_omnibench"
    data_object = datasets.renku_dataset_create(
        name=data_name, kg_url="https://renkulab.io/knowledge-graph"
    )
    remove_dataset_command().build().execute("test_omnibench")

    assert data_object.name == data_name


def test_renku_dataset_create_no_project_context(no_project_context):

    with pytest.raises(ProjectError, match=r"Directory is not a renku project*?"):
        datasets.renku_dataset_create(name="mock_dataset", kg_url="some.url")


### Test renku_dataset_import
def test_renku_dataset_import_no_project_context(no_project_context):

    with pytest.raises(ProjectError, match=r"Directory is not a renku project*?"):
        datasets.renku_dataset_import(uri="some.url")


@pytest.mark.renku_call
def test_renku_dataset_import_works():

    data_object = datasets.renku_dataset_import(
        uri="https://renkulab.io/datasets/94638c493af44b5e8bc09b2351f6a8c2",
        name="test_dataset",
    )
    assert os.path.exists("data/test_dataset/data_info_cellbench.json")
    shutil.rmtree("data/test_dataset", ignore_errors=True)
    remove_dataset_command().build().execute("test_dataset")


### Test renku_dataset_update
def test_renku_dataset_update_no_project_context(no_project_context):

    with pytest.raises(ProjectError, match=r"Directory is not a renku project*?"):
        datasets.renku_dataset_update(names=["some_dataset", "another_datatset"])


def test_renku_dataset_update_no_input(monkeypatch):
    def get_mock_status():
        return True

    monkeypatch.setattr(
        omnibenchmark.management.general_checks,
        "is_renku_project",
        lambda *args, **kwargs: get_mock_status(),
    )
    with pytest.raises(errors.ParameterError, match=r"Either names, update_all*?"):
        datasets.renku_dataset_update(names=[])


def test_renku_dataset_update_ambigous_inputs(monkeypatch):
    def get_mock_status():
        return True

    monkeypatch.setattr(
        omnibenchmark.management.general_checks,
        "is_renku_project",
        lambda *args, **kwargs: get_mock_status(),
    )

    with pytest.raises(
        errors.ParameterError, match=r"Cannot pass dataset names with update_all"
    ):
        datasets.renku_dataset_update(names=["test2", "something"], update_all=True)


def test_renku_dataset_update_ambigous_inputs2(monkeypatch):
    def get_mock_status():
        return True

    monkeypatch.setattr(
        omnibenchmark.management.general_checks,
        "is_renku_project",
        lambda *args, **kwargs: get_mock_status(),
    )

    with pytest.raises(
        errors.ParameterError, match=r"Cannot pass include/exclude with update_all"
    ):
        datasets.renku_dataset_update(
            names=[], include=["test2", "something"], update_all=True
        )


@pytest.mark.renku_call
def test_renku_dataset_update_works():

    res = datasets.renku_dataset_update(names=[], update_all=True)
    assert res is None


### Test renku_add_to_dataset
def test_renku_add_to_dataset_no_project_context(no_project_context):

    with pytest.raises(ProjectError, match=r"Directory is not a renku project*?"):
        datasets.renku_add_to_dataset(
            urls=["some_dataset", "another_datatset"], dataset_name="some_dataset"
        )


@pytest.mark.renku_call
def test_renku_add_to_dataset_works():
    test_file = "data/utils_test/test_file.txt"
    text_file = open(test_file, "w")
    text_file.write("This is a test for renku dataset add")
    text_file.close()

    res = datasets.renku_add_to_dataset(urls=[test_file], dataset_name="utils_test")
    datasets = Dataset.list()
    data_files = [dataset.files for dataset in datasets if dataset.name == "utils_test"]
    assert test_file in data_files
    file_unlink_command().build().execute(name="utils_test", include=test_file)



"""Checks concerning (usually renku) datasets"""

from renku.ui.api.models.dataset import Dataset
import requests
from typing import Union
import os


def renku_dataset_exist(name: str, path: Union[os.PathLike, str] = os.getcwd()) -> bool:
    """Check if a renku dataset with a specific name already exist at a certain project path.

    Args:
        name (str): Dataset name
        path (PathLike, optional): Project path to check. Defaults to ".".

    Returns:
        bool: True/False a dataset with that name exist in the project at that path.
    """

    current_dir = os.getcwd()
    os.chdir(path)
    datasets = Dataset.list()
    matches = [dataset.name for dataset in datasets if dataset.name == name]
    os.chdir(current_dir)
    return True if len(matches) >= 1 else False


def dataset_name_exist(name: str, kg_url: str) -> bool:
    """Check if a renku dataset with a defined name already exists in the knowledge base.

    Args:
        name (str): Name to query
        kg_url (str): Url of the knowledge base to query

    Returns:
        bool: True/False, if a dataset with that name already exist
    """

    url = kg_url + "/datasets?query=" + name
    response = requests.get(url)
    return True if name in [item["name"] for item in response.json()] else False



"""General checks to ensure the enviroment and setup are as expected"""

from renku.ui.api.models.project import Project
from omnibenchmark.utils.exceptions import InputError
from typing import Union, Optional
import os
import lxml.html as lh
import requests


def is_renku_project(path: Union[os.PathLike, str] = os.getcwd()) -> bool:
    """Checks if a project is an initialized renku project.
    Args:
        path (Pathlike, str, optional): A path to check. Defaults to ".".

    Returns:
        bool: True if project is a renku project.
    """
    current_dir = os.getcwd()
    os.chdir(path)
    project = Project()
    repo = project.client.repository
    os.chdir(current_dir)
    return True if repo.contains(".renku/metadata") else False


def find_orchestrator(
    benchmark_name: str,
    bench_url: str = "https://omnibenchmark.pages.uzh.ch/omni_dash/benchmarks",
    key_header: str = "Benchmark_name",
    o_header: str = "Orchestrator",
) -> Optional[str]:
    bench_html = requests.get(bench_url)
    doc = lh.fromstring(bench_html.content)
    tr_elements = doc.xpath("//tr")
    header = tr_elements[0]
    col_num = 0
    for field in header:
        if field.text_content() == key_header:
            nam_col = col_num
        if field.text_content() == o_header:
            o_col = col_num
        col_num += 1

    try:
        nam_col, o_col
    except Exception as e:
        raise InputError(
            f"Could not find columns with names {key_header} and/or {o_header}.\n"
            f"Please check {bench_url} for the correct column names."
        )

    o_obj = [
        tr_ele[o_col]
        for tr_ele in tr_elements
        if benchmark_name in tr_ele[nam_col].text_content().split(",")
    ]
    if len(o_obj) < 1:
        print(
            f"WARNING: Could not find benchmark associated to {benchmark_name}.\n"
            f"Check {bench_url} for existing benchmarks.\n"
            f"Integration with existing projects is not possible."
        )
        return None
    o_url = o_obj[0].text_content()
    return o_url



"""Classes related to omnibenchmark outputs"""

from typing import Mapping, List, Optional, Union, Callable
from omnibenchmark.utils.user_input_checks import check_name_matching, flatten
from omnibenchmark.utils.exceptions import InputError, OutputError
from omnibenchmark.utils.auto_output import (
    get_all_output_combinations,
    get_default_outputs,
    autocomplete_file_mapping,
)
from omnibenchmark.utils.auto_run import map_plan_names_file_types
from omnibenchmark.core.input_classes import OmniInput, OmniParameter, OutMapping
from omnibenchmark.utils.auto_command import (
    automatic_command_generation,
    get_interpreter_from_extension,
)
from renku.command.view_model.plan import PlanViewModel
from os import PathLike
import os


class OmniOutput:
    def __init__(
        self,
        name: str,
        out_names: List[str],
        output_end: Optional[Mapping[str, str]] = None,
        out_template: str = "data/${name}/${name}_${unique_values}_${out_name}.${out_end}",
        file_mapping: Optional[List[OutMapping]] = None,
        inputs: Optional[OmniInput] = None,
        parameter: Optional[OmniParameter] = None,
        default: Optional[Mapping] = None,
        template_fun: Optional[Callable[..., Mapping]] = None,
        template_vars: Optional[Mapping] = None,
    ):
        self.name = name
        self.out_names = out_names
        self.output_end = output_end
        self.out_template = out_template
        self.file_mapping = file_mapping
        self.inputs = inputs
        self.parameter = parameter
        self.default = default
        self.template_fun = template_fun
        self.template_vars = template_vars if template_vars is not None else {}

        if self.file_mapping is None:

            if self.output_end is None:
                raise OutputError(
                    f"Undefined outputs. Please specify an output file mapping or output prefix and output ending to automatically generate one."
                )

            check_name_matching(self.out_names, self.output_end.keys())
            self.file_mapping = get_all_output_combinations(  
                name=self.name,
                output_end=self.output_end,
                out_template=self.out_template,
                inputs=self.inputs,
                parameter=self.parameter,
                template_fun=self.template_fun,
                **self.template_vars,
            )
        
        self.file_mapping = autocomplete_file_mapping(self.file_mapping)

        check_name_matching(
            self.out_names,
            flatten(
                [
                    out_mapping["output_files"].keys()  # type: ignore
                    for out_mapping in self.file_mapping
                ]
            ),
        )

        if self.default is None:
            self.default = get_default_outputs(
                file_mapping=self.file_mapping,
                inputs=self.inputs,
                parameter=self.parameter,
            )
        if self.default is not None:
            check_name_matching(self.out_names, self.default.keys())

    def update_outputs(self):
        if self.inputs is not None or self.parameter is not None:
            self.template_vars = self.template_vars if self.template_vars is not None else {}
            new_file_mapping = get_all_output_combinations( 
                name=self.name,
                output_end=self.output_end,
                out_template=self.out_template,
                inputs=self.inputs,
                parameter=self.parameter,
                template_fun=self.template_fun,
                **self.template_vars,
            )
            ex_file_mapping = self.file_mapping if self.file_mapping is not None else []
            all_file_mappings = ex_file_mapping + new_file_mapping
            self.file_mapping = list({str(i):i for i in all_file_mappings}.values())

        check_name_matching(
            self.out_names,
            flatten(
                [
                    out_mapping["output_files"].keys()  # type: ignore
                    for out_mapping in self.file_mapping
                ]
            ),
        )

        if self.default is None:
            self.default = get_default_outputs(
                file_mapping=self.file_mapping,
                inputs=self.inputs,
                parameter=self.parameter,
            )
        #return self


class OmniCommand:
    def __init__(
        self,
        script: Union[PathLike, str],
        interpreter: str = None,
        command_line: str = None,
        outputs: OmniOutput = None,
    ):
        self.script = script
        self.interpreter = interpreter
        self.command_line = command_line
        self.outputs = outputs

        if self.command_line is None:
            if self.outputs is None or self.outputs.file_mapping is None:
                raise InputError(
                    f'Can not infer command without Output definition. Please specify either "command_line" or "outputs".'
                )
            output_val = self.outputs.default
            input_val = next(
                (
                    out_dict["input_files"]
                    for out_dict in self.outputs.file_mapping
                    if out_dict["output_files"] == output_val
                ),
                None,
            )
            parameter_val = next(
                (
                    out_dict["parameter"]
                    for out_dict in self.outputs.file_mapping
                    if out_dict["output_files"] == output_val
                ),
                None,
            )
            if self.interpreter is None:
                ext = os.path.splitext(script)[1]
                self.interpreter = get_interpreter_from_extension(ext)

            self.command_line = automatic_command_generation(
                self.script,
                interpreter=self.interpreter,
                inputs=input_val,
                outputs=output_val,
                parameters=parameter_val,
            )
        ## Add command checks!


class OmniPlan:
    def __init__(
        self, plan: PlanViewModel, param_mapping: Optional[Mapping[str, str]] = None
    ):
        self.plan = plan
        self.param_mapping = param_mapping

    def predict_mapping_from_file_dict(
        self, file_dict=Mapping[str, str]
    ) -> Mapping[str, str]:
        return map_plan_names_file_types(plan=self.plan, file_dict=file_dict)



from omnibenchmark.core.output_classes import OmniOutput
from omnibenchmark.utils.exceptions import OutputError, InputError
import pytest


### Test OmniOutput class
def test_omni_output_with_output_mapping(mock_out_mapping):
    test_output = OmniOutput(
        name="mock_res",
        out_names=["out_file1", "out_file2"],
        file_mapping=[mock_out_mapping],
    )
    assert isinstance(test_output, OmniOutput)


def test_omni_output_with_out_mapping_not_match_names(mock_out_mapping):
    with pytest.raises(InputError, match=r"Expected inputs from .*?"):
        OmniOutput(
            name="mock_res",
            out_names=["quatsch", "out_file2"],
            file_mapping=[mock_out_mapping],
        )


def test_omni_output_without_outmapping_out_end():
    with pytest.raises(
        OutputError, match=r"Undefined outputs. Please specify an output file .*?"
    ):
        OmniOutput(name="mock_res", out_names=["quatsch", "out_file2"])


def test_omni_output_without_mapping_not_match_outend_names():
    with pytest.raises(InputError, match=r"Expected inputs from .*?"):
        OmniOutput(
            name="mock_res",
            out_names=["out_file1", "out_file2"],
            output_end={"out1": ".txt", "out_file2": ".gz"},
        )


def test_omni_output_without_mapping_works():
    test_output = OmniOutput(
        name="mock_res",
        out_names=["out_file1", "out_file2"],
        output_end={"out_file1": ".txt", "out_file2": ".gz"},
    )
    assert isinstance(test_output, OmniOutput)


def test_omni_output_automatic_default(mock_out_mapping):
    test_output = OmniOutput(
        name="mock_res",
        out_names=["out_file1", "out_file2"],
        file_mapping=[mock_out_mapping],
    )
    assert test_output.default == mock_out_mapping["output_files"]


def test_omni_output_default_name_mismatch(mock_out_mapping):
    with pytest.raises(InputError, match=r"Expected inputs from .*?"):
        mis_match = {"out_file1": "path/to/out1", "out_fileX": "path/to/out2"}
        OmniOutput(
            name="mock_res",
            out_names=["out_file1", "out_file2"],
            file_mapping=[mock_out_mapping],
            default=mis_match,
        )



from typing import Optional, List
from renku.command.save import save_and_push_command


def renku_save(
    message: Optional[str] = None,
    destination: Optional[str] = None,
    paths: Optional[List[str]] = None,
):

    saved_paths, branch = (
        save_and_push_command()
        .build()
        .execute(message=message, remote=destination, paths=paths)
        .output
    )
    if saved_paths:
        print(
            "Successfully saved to remote branch {}: \n\t{}".format(
                branch, "\n\t".join(saved_paths)
            )
        )
    else:
        print("There were no changes to save.")
